"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""

# NB: can_read, can_write imply that *some* firmware versions support them.
#     Checks with feature_flag are important, especially for earlier firmware.

import asyncio
import copy
import enum
import time
import traceback
from struct import unpack, pack
from typing import Union, Optional

import pyDE1
from pyDE1.de1.ble import CUUID
from pyDE1.exceptions import (
    DE1APITypeError, DE1APIValueError, DE1APITooManyFramesError,
    MMRTypeError, MMRValueError, MMRDataTooLongError
)
from pyDE1.utils import data_as_hex

logger = pyDE1.getLogger('DE1.C_API')

# TODO: log_string() is not "None-safe"
#       TypeError: unsupported format string passed to NoneType.__format__


#
# Utilities for range check of APi types
#

def validate_f8_1_7(value):
    """
    Range check of F8_1_7 data type which is
    an unsigned, 7-bit mantissa
    with an exponent of either 0.1 or 1
    """
    if value < 0 or value > 127.5:
        raise DE1APIValueError("Out of range for F8_1_7 data type")
    return value


def validate_f8_1_7_noneok(value):
    if value is None:
        return None
    return validate_f8_1_7(value)


def f8_1_7_decode(value):
    mantissa = value & 0x7f
    if (value & 0x80):
        exponent = 1
    else:
        exponent = 0.1
    return mantissa * exponent


def f8_1_7(value):
    if value < 0:
        raise DE1APIValueError("Out of range for F8_1_7 data type")
    elif value < 12.75:
        retval = int(round(value * 10))
    elif value < 127.5:
        retval = int(round(value)) + 0x80
    else:
        raise DE1APIValueError("Out of range for F8_1_7 data type")
    return retval


def validate_s_p(value, s, p):
    scaled = round(value / 2**p)
    maxabs = 2**(s - 1)
    if scaled < -maxabs or scaled > (maxabs - 1):
        raise DE1APIValueError(f"Out of range for S{s}P{p} data type")
    return value

def validate_s_p_noneok(value, s, p):
    if value is None:
        return None
    return validate_s_p(value, s, p)


def validate_u_p(value, u, p):
    scaled = round(value / 2**p)
    if scaled < 0 or scaled > (2**u - 1):
        raise DE1APIValueError(f"Out of range for U{u}P{p} data type")
    return value


def validate_u_p_noneok(value, u, p):
    if value is None:
        return None
    return validate_u_p(value, u, p)

def u(val):
    return int(round(val))

def p0(val):
    return int(round(val))

def p1(val):
    return int(round(val * 2))

def p4(val):
    return int(round(val * 16))

def p8(val):
    return int(round(val * 256))

def p12(val):
    return int(round(val * 4096))

def p16(val):
    return int(round(val * 65536))


class PackedAttr:
    """
    Abstract parent of classes used represent struct PACKEDATTR
    from APIDataTypes.hpp (code running on the DE1)

    All should be able to support serialization and deserialization
    as well as providing a "log-safe" representation string.

    Range checking is generally done in subclasses. Can raise:
    DE1APITypeError and DE1APIValueError and subclasses
    """

    def __init__(self):
        self._arrival_time = None
        pass

    def from_wire_bytes(self, wire_bytes: Union[bytes, bytearray], arrival_time=None):
        self._arrival_time = arrival_time

        return self

    def as_wire_bytes(self) -> Union[bytes, bytearray]:
        raise NotImplementedError

    def log_string(self):
        return self.__repr__()

    @property
    def arrival_time(self):
        return self._arrival_time

    # Boilerplate here (could all be False for base class with cuuid = None)
    cuuid = None
    can_read = bool(cuuid is not None and cuuid.can_read)
    can_write = bool(cuuid is not None and cuuid.can_write)
    can_notify = bool(cuuid is not None and cuuid.can_notify)
    can_write_then_return = bool(cuuid is not None
                                 and cuuid.can_write_then_return)



def get_cuuid(packed_attr: PackedAttr) -> CUUID:
    """
    :Protecting" getter -- raises if None
    """
    if (cuuid := packed_attr.cuuid) is None:
        raise DE1APIValueError(
            f"Not an over-the-wire PackedAttr {packed_attr}"
        )
    return cuuid



# def packed_attr_from_cuuid(cuuid: CUUID) -> PackedAttr:
# found at the end, as derived at load time from list of subclasses


#
# Follow the order in APUDataTypes.hpp for ease of maintenance
#


# Present, but apparently unused
class Models (enum.IntEnum):
    Model_Plus    = 1
    Model_Pro     = 2
    Model_CAFE    = 4


# Present, but apparently unused
class Config (enum.IntFlag):
    LowPowerHeater  = 0x1
    CFG_Refill      = 0x2
    Voltage_120V    = 0x4


class FWVersion (PackedAttr):

    def __init__(self, APIVersion=None, Release=None, Commits=None,
                 Changes=None, BLESha=None):
        super(FWVersion, self).__init__( )
        self.APIVersion = APIVersion
        self.Release = Release
        self.Commits = Commits
        self.Changes = Changes
        self.BLESha = BLESha

    def from_wire_bytes(self, wire_bytes, arrival_time=None):
        super(FWVersion, self).from_wire_bytes(wire_bytes, arrival_time)
        (
            self._APIVersion,
            self._Release,  # F8_1_7
            self._Commits,
            self._Changes,
            self._BLESha,  # TODO: Interpret as 7 nibbles + dirty
         ) = unpack('>BBHBL', wire_bytes)
        self._Release = f8_1_7_decode(self._Release)

        return self

    def as_wire_bytes(self):
        return pack('>BBHBL',
                    p0(self._APIVersion),
                    f8_1_7(self._Release),
                    p0(self._Commits),
                    p0(self._Changes),
                    p0(self._BLESha),
                    )

    def log_string(self):
        return 'API: {} Release: {:.2f} Commits: {} Changes: {} SHA: {:07x}({})'.format(
            self._APIVersion,
            self._Release,
            self._Commits,
            self._Changes,
            self._BLESha >> 4,
            self._BLESha & 0x0f
        )

    @property
    def APIVersion(self):
        return self._APIVersion

    @APIVersion.setter
    def APIVersion(self, value):
        self._APIVersion = validate_u_p_noneok(value, 8, 0)

    @property
    def Release(self):
        return self._Release

    @Release.setter
    def Release(self, value):
        self._Release = validate_f8_1_7_noneok(value)

    @property
    def Commits(self):
        return self._Commits

    @Commits.setter
    def Commits(self, value):
        self._Commits = validate_u_p_noneok(value, 16, 0)

    @property
    def Changes(self):
        return self._Changes

    @Changes.setter
    def Changes(self, value):
        self._Changes = validate_u_p_noneok(value, 8, 0)

    @property
    def BLESha(self):
        return self._BLESha

    @BLESha.setter
    def BLESha(self, value):
        self._BLESha = validate_u_p_noneok(value, 32, 0)

    def BLESha_is_dirty(self):
        return (self._BLESha & 0b1 == 0b1)


class Versions (PackedAttr):

    cuuid = CUUID.Versions
    can_read = bool(cuuid is not None and cuuid.can_read)
    can_write = bool(cuuid is not None and cuuid.can_write)
    can_notify = bool(cuuid is not None and cuuid.can_notify)
    can_write_then_return = bool(cuuid is not None
                                 and cuuid.can_write_then_return)

    def __init__(self, BLEVersion=None, LVVersion=None):
        super(Versions, self).__init__()

        self.BLEVersion = BLEVersion
        self.LVVersion = LVVersion


    def from_wire_bytes(self, wire_bytes, arrival_time=None):
        super(Versions, self).from_wire_bytes(wire_bytes, arrival_time)
        self._BLEVersion.from_wire_bytes(wire_bytes[0:9], arrival_time)
        self._LVVersion.from_wire_bytes(wire_bytes[9:], arrival_time)

        return self

    def as_wire_bytes(self):
        retval = bytearray(self._BLEVersion.as_wire_bytes())
        retval.extend(self._LVVersion.as_wire_bytes())
        return retval

    def log_string(self):
        return 'BLE: {} LV: {}'.format(
            self._BLEVersion.log_string(),
            self._LVVersion.log_string(),
        )

    @property
    def BLEVersion(self):
        return copy.deepcopy(self._BLEVersion)

    @BLEVersion.setter
    def BLEVersion(self, obj):
        if not (isinstance(obj, FWVersion) or obj is None):
            raise DE1APITypeError("Expecting FWVersion")
        if obj is None:
            obj = FWVersion()
        self._BLEVersion = obj

    @property
    def LVVersion(self):
        return copy.deepcopy(self._LVVersion)

    @LVVersion.setter
    def LVVersion(self, obj):
        if not (isinstance(obj, FWVersion) or obj is None):
            raise DE1APITypeError("Expecting FWVersion")
        if obj is None:
            obj = FWVersion()
        self._LVVersion = obj


# Present, but apparently unused
class BoardVersion (enum.IntEnum):
    PRE_H   = 0
    H       = 1


class TemperatureSet (PackedAttr):

    def __init__(self, WaterHeater=None, SteamHeater=None,
                 GroupHeater=None, ColdWater=None):
        super(TemperatureSet, self).__init__()
        self.WaterHeater = WaterHeater
        self.SteamHeater = SteamHeater
        self.GroupHeater = GroupHeater
        self.ColdWater = ColdWater

    def from_wire_bytes(self, wire_bytes, arrival_time=None):
        super(TemperatureSet, self).from_wire_bytes(wire_bytes, arrival_time)
        (
            self._WaterHeater,
            self._SteamHeater,
            self._GroupHeater,
            self._ColdWater
         ) = unpack(">HHHH", wire_bytes)

        self._WaterHeater = self._WaterHeater / 2**8
        self._SteamHeater = self._SteamHeater / 2**8
        self._GroupHeater = self._GroupHeater / 2**8
        self._ColdWater = self._ColdWater / 2**8

        return self

    def as_wire_bytes(self):
        raise NotImplementedError

    def log_string(self):
        return 'Water: {:.2f} Steam: {:.2f} Group: {:.2f} Cold: {:.2f}'.format(
            self._WaterHeater,
            self._SteamHeater,
            self._GroupHeater,
            self._ColdWater,
        )

    @property
    def WaterHeater(self):
        return self._WaterHeater

    @WaterHeater.setter
    def WaterHeater(self, value):
        self._WaterHeater = validate_u_p_noneok(value, 16, 8)

    @property
    def SteamHeater(self):
        return self._SteamHeater

    @SteamHeater.setter
    def SteamHeater(self, value):
        self._SteamHeater = validate_u_p_noneok(value, 16, 8)

    @property
    def GroupHeater(self):
        return self._GroupHeater

    @GroupHeater.setter
    def GroupHeater(self, value):
        self._GroupHeater = validate_u_p_noneok(value, 16, 8)

    @property
    def ColdWater(self):
        return self._ColdWater

    @ColdWater.setter
    def ColdWater(self, value):
        self._ColdWater = validate_u_p_noneok(value, 16, 8)


class Temperatures (PackedAttr):

    cuuid = CUUID.Temperatures
    can_read = bool(cuuid is not None and cuuid.can_read)
    can_write = bool(cuuid is not None and cuuid.can_write)
    can_notify = bool(cuuid is not None and cuuid.can_notify)
    can_write_then_return = bool(cuuid is not None
                                 and cuuid.can_write_then_return)

    def __init__(self, Current=None, Target=None):
        super(Temperatures, self).__init__()

        self.Current = Current
        self.Target = Target


    def from_wire_bytes(self, wire_bytes, arrival_time=None):
        super(Temperatures, self).from_wire_bytes(wire_bytes, arrival_time)
        self._Current.from_wire_bytes(wire_bytes[0:8], arrival_time)
        self._Target.from_wire_bytes(wire_bytes[8:], arrival_time)

        return self

    def as_wire_bytes(self):
        raise NotImplementedError

    def log_string(self):
        return 'Current: {} Target: {}'.format(
            self._Current.log_string(),
            self._Target.log_string(),
        )

    @property
    def Current(self):
        return copy.deepcopy(self._Current)

    @Current.setter
    def Current(self, obj):
        if not (isinstance(obj, TemperatureSet) or obj is None):
            raise DE1APITypeError("Expecting TemperatureSet")
        if obj is None:
            obj = TemperatureSet()
        self._Current = obj

    @property
    def Target(self):
        return copy.deepcopy(self._Target)

    @Target.setter
    def Target(self, obj):
        if not (isinstance(obj, TemperatureSet) or obj is None):
            raise DE1APITypeError("Expecting TemperatureSet")
        if obj is None:
            obj = TemperatureSet()
        self._Target = obj


class SteamSetting (enum.IntFlag):
    NoneSet   = 0x00  # Otherwise returns SlowStart
    FastStart = 0x80
    SlowStart = 0x00
    HighPower = 0x40
    LowPower  = 0x00


class ShotSettings (PackedAttr):

    cuuid = CUUID.ShotSettings
    can_read = bool(cuuid is not None and cuuid.can_read)
    can_write = bool(cuuid is not None and cuuid.can_write)
    can_notify = bool(cuuid is not None and cuuid.can_notify)
    can_write_then_return = bool(cuuid is not None
                                 and cuuid.can_write_then_return)

    def __init__(self, SteamSettings=None,
                 TargetSteamTemp=None, TargetSteamLength=None,
                 TargetHotWaterTemp=None, TargetHotWaterVol=None,
                 TargetHotWaterLength=None, TargetEspressoVol=None,
                 TargetGroupTemp=None
                 ):
        super(ShotSettings, self).__init__()

        self.SteamSettings = SteamSettings
        self.TargetSteamTemp = TargetSteamTemp
        self.TargetSteamLength = TargetSteamLength
        self.TargetHotWaterTemp = TargetHotWaterTemp
        self.TargetHotWaterVol = TargetHotWaterVol
        self.TargetSteamLength = TargetHotWaterLength
        self.TargetEspressoVol = TargetEspressoVol
        self.TargetGroupTemp = TargetGroupTemp


    def from_wire_bytes(self, wire_bytes, arrival_time=None):
        super(ShotSettings, self).from_wire_bytes(wire_bytes, arrival_time)
        (self._SteamSettings,
        self._TargetSteamTemp,
        self._TargetSteamLength,
        self._TargetHotWaterTemp,
        self._TargetHotWaterVol,
        self._TargetHotWaterLength,
        self._TargetEspressoVol,
        self._TargetGroupTemp,
         ) = unpack('>BBBBBBBH', wire_bytes)

        self._TargetGroupTemp = self._TargetGroupTemp / 2**8

        return self

    def as_wire_bytes(self):
        return pack('>BBBBBBBH',
            p0(self._SteamSettings),
            p0(self._TargetSteamTemp),
            p0(self._TargetSteamLength),
            p0(self._TargetHotWaterTemp),
            p0(self._TargetHotWaterVol),
            p0(self._TargetHotWaterLength),
            p0(self._TargetEspressoVol),
            p8(self._TargetGroupTemp),
        )

    def log_string(self):
        if self._SteamSettings & SteamSetting.FastStart.value:
            steam_settings = "Fast"
        else:
            steam_settings = "Slow"
        if self._SteamSettings & SteamSetting.HighPower.value:
            steam_settings += ",High"
        else:
            steam_settings += ",Low"
        return 'Steam: {} Temp: {} Length: {}; ' \
            'Water: Temp: {} Vol: {} Length: {}; ' \
            'EVol: {} GTemp: {}'.format(
            steam_settings,
            self._TargetSteamTemp,
            self._TargetSteamLength,
            self._TargetHotWaterTemp,
            self._TargetHotWaterVol,
            self._TargetHotWaterLength,
            self._TargetEspressoVol,
            self._TargetGroupTemp,
        )

    @property
    def SteamSettings(self):
        return SteamSetting(self._SteamSettings)

    @SteamSettings.setter
    def SteamSettings(self, value):
        self._SteamSettings = validate_u_p_noneok(value, 8, 0)

    @property
    def steam_setting_fast_start(self):
        return bool(self.SteamSettings & SteamSetting.FastStart)

    @steam_setting_fast_start.setter
    def steam_setting_fast_start(self, val: bool):
        if val:
            self.SteamSettings = self.SteamSettings | SteamSetting.FastStart
        else:
            self.SteamSettings = self.SteamSettings & ~ SteamSetting.FastStart

    @property
    def steam_setting_high_power(self):
        return bool(self.SteamSettings & SteamSetting.FastStart)

    @steam_setting_high_power.setter
    def steam_setting_high_power(self, val: bool):
        if val:
            self.SteamSettings = self.SteamSettings | SteamSetting.HighPower
        else:
            self.SteamSettings = self.SteamSettings ^ SteamSetting.HighPower

    @property
    def TargetSteamTemp(self):
        return self._TargetSteamTemp

    @TargetSteamTemp.setter
    def TargetSteamTemp(self, value):
        if value is not None and (value < 140 or 160 < value) :
            raise DE1APIValueError("TargetSteamTemp must be 140 - 160")
        self._TargetSteamTemp = validate_u_p_noneok(value, 8, 0)

    @property
    def TargetSteamLength(self):
        return self._TargetSteamLength

    @TargetSteamLength.setter
    def TargetSteamLength(self, value):
        self._TargetSteamLength = validate_u_p_noneok(value, 8, 0)

    @property
    def TargetHotWaterTemp(self):
        return self._TargetHotWaterTemp

    @TargetHotWaterTemp.setter
    def TargetHotWaterTemp(self, value):
        self._TargetHotWaterTemp = validate_u_p_noneok(value, 8, 0)

    @property
    def TargetHotWaterVol(self):
        return self._TargetHotWaterVol

    @TargetHotWaterVol.setter
    def TargetHotWaterVol(self, value):
        self._TargetHotWaterVol = validate_u_p_noneok(value, 8, 0)

    @property
    def TargetHotWaterLength(self):
        return self._TargetHotWaterLength

    @TargetHotWaterLength.setter
    def TargetHotWaterLength(self, value):
        self._TargetHotWaterLength = validate_u_p_noneok(value, 8, 0)

    @property
    def TargetEspressoVol(self):
        return self._TargetEspressoVol

    @TargetEspressoVol.setter
    def TargetEspressoVol(self, value):
        self._TargetEspressoVol = validate_u_p_noneok(value, 8, 0)

    @property
    def TargetGroupTemp(self):
        return self._TargetGroupTemp

    @TargetGroupTemp.setter
    def TargetGroupTemp(self, value):
        self._TargetGroupTemp = validate_u_p_noneok(value, 16, 8)


class API_MachineStates (enum.IntEnum):
    Sleep           = 0x00
    GoingToSleep    = 0x01
    Idle            = 0x02
    Busy            = 0x03
    Espresso        = 0x04
    Steam           = 0x05
    HotWater        = 0x06
    ShortCal        = 0x07
    SelfTest        = 0x08
    LongCal         = 0x09
    Descale         = 0x0a
    FatalError      = 0x0b
    Init            = 0x0c
    NoRequest       = 0x0d  # Allows RequestedState to be sent as a noop
    SkipToNext      = 0x0e
    HotWaterRinse   = 0x0f
    SteamRinse      = 0x10
    Refill          = 0x11
    Clean           = 0x12
    InBootLoader    = 0x13
    AirPurge        = 0x14
    SchedIdle       = 0x15

    @property
    def is_flow_state(self):
        return self in (
            self.Espresso,
            self.Steam,
            self.HotWater,
            self.HotWaterRinse
        )


class API_Substates (enum.IntEnum):
    NoState             = 0x00
    HeatWaterTank       = 0x01
    HeatWaterHeater     = 0x02
    StabilizeMixTemp    = 0x03
    PreInfuse           = 0x04
    Pour                = 0x05
    Flush               = 0x06
    Steaming            = 0x07
    DescaleInit         = 0x08
    DescaleFillGroup    = 0x09
    DescaleReturn       = 0x0a
    DescaleGroup        = 0x0b
    DescaleSteam        = 0x0c
    CleanInit           = 0x0d
    CleanFillGroup      = 0x0e
    CleanSoak           = 0x0f
    CleanGroup          = 0x10
    PausedRefill        = 0x11
    PausedSteam         = 0x12

    Error_NaN           = 200
    Error_Inf           = 201
    Error_Generic       = 202
    Error_ACC           = 203
    Error_TSensor       = 204
    Error_PSensor       = 205
    Error_WLevel        = 206
    Error_DIP           = 207
    Error_Assertion     = 208
    Error_Unsafe        = 209
    Error_InvalidParm   = 210
    Error_Flash         = 211
    Error_OOM           = 212
    Error_Deadline      = 213
    Error_HiCurrent     = 214
    Error_LoCurrent     = 215
    Error_BootFill      = 216

    @property
    def is_error(self):
        return self.value >= 200

    @property
    def flow_phase(self):
        if self in (
            self.HeatWaterTank,
            self.HeatWaterHeater,
            self.StabilizeMixTemp
        ):
            retval = 'before'
        elif self in (
            self.PreInfuse,
            self.Pour,
            self.Steaming,
            self.PausedSteam
        ):
            retval = 'during'
        elif self in (
            self.Flush,
        ):
            retval = 'after'
        else:
            retval = None
        return retval

class StateInfo (PackedAttr):

    cuuid = CUUID.StateInfo
    can_read = bool(cuuid is not None and cuuid.can_read)
    can_write = bool(cuuid is not None and cuuid.can_write)
    can_notify = bool(cuuid is not None and cuuid.can_notify)
    can_write_then_return = bool(cuuid is not None
                                 and cuuid.can_write_then_return)

    def __init__(self, State=None, SubState=None):
        super(StateInfo, self).__init__()

        self.State = State
        self.SubState = SubState


    def from_wire_bytes(self, wire_bytes, arrival_time=None):
        super(StateInfo, self).from_wire_bytes(wire_bytes, arrival_time)
        # Even though "trivial", use of unpack checks against format
        ( state, substate ) = unpack('>BB', wire_bytes)
        self._State = API_MachineStates(state)
        self._SubState = API_Substates(substate)

        return self

    def as_wire_bytes(self):
        raise NotImplementedError

    def log_string(self):
        return '{},{}'.format(
            self._State.name,
            self._SubState.name,
        )

    @property
    def State(self):
        return self._State

    @State.setter
    def State(self, value):
        if not (isinstance(value, API_MachineStates) or value is None):
            raise DE1APITypeError("Expecting API_MachineStates")
        self._State = value

    @property
    def SubState(self):
        return self._SubState

    @SubState.setter
    def SubState(self, value):
        if not (isinstance(value, API_Substates) or value is None):
            raise DE1APITypeError("Expecting API_Substates")
        self._SubState = value


class RequestedState (PackedAttr):

    cuuid = CUUID.RequestedState
    can_read = bool(cuuid is not None and cuuid.can_read)
    can_write = bool(cuuid is not None and cuuid.can_write)
    can_notify = bool(cuuid is not None and cuuid.can_notify)
    can_write_then_return = bool(cuuid is not None
                                 and cuuid.can_write_then_return)

    # NB: struct member is RequestedState
    def __init__(self, State=None):
        super(RequestedState, self).__init__()

        self.RequestedState = State


    def as_wire_bytes(self):
        return pack('>B', u(self._RequestedState.value))

    def from_wire_bytes(self, wire_bytes, arrival_time=None):
        super(RequestedState, self).from_wire_bytes(wire_bytes, arrival_time)
        state = unpack('>B', wire_bytes)[0]
        self._RequestedState = API_MachineStates(state)

        return self

    def log_string(self):
        return self._RequestedState.name

    @property
    def RequestedState(self):
        return self._RequestedState

    @RequestedState.setter
    def RequestedState(self, value):
        if not (isinstance(value, API_MachineStates) or value is None):
            raise DE1APITypeError("Expecting API_MachineStates")
        self._RequestedState = value


class WaterLevels (PackedAttr):

    cuuid = CUUID.WaterLevels
    can_read = bool(cuuid is not None and cuuid.can_read)
    can_write = bool(cuuid is not None and cuuid.can_write)
    can_notify = bool(cuuid is not None and cuuid.can_notify)
    can_write_then_return = bool(cuuid is not None
                                 and cuuid.can_write_then_return)

    def __init__(self, Level=None, StartFillLevel=None):
        super(WaterLevels, self).__init__()

        self.Level = Level
        self.StartFillLevel = StartFillLevel


    def from_wire_bytes(self, wire_bytes, arrival_time=None):
        super(WaterLevels, self).from_wire_bytes(wire_bytes, arrival_time)
        (level, start_fill_level) = unpack('>HH', wire_bytes)
        self._Level = level / 2**8
        self._StartFillLevel = start_fill_level / 2**8

        return self

    def as_wire_bytes(self):
        level = self._Level
        if level is None:
            level = 0
        start_fill_level = self._StartFillLevel
        if start_fill_level is None:
            start_fill_level = 0
        return pack('>HH',
                    u(level),
                    u(start_fill_level)
                    )

    def log_string(self):
        return 'Level: {:.2f} Refill: {:.2f}'.format(
            self._Level,
            self._StartFillLevel,
        )

    @property
    def Level(self):
        return self._Level

    @Level.setter
    def Level(self, value):
        self._Level = validate_u_p_noneok(value, 16,8)

    @property
    def StartFillLevel(self):
        return self._StartFillLevel

    @StartFillLevel.setter
    def StartFillLevel(self, value):
        self._StartFillLevel = validate_u_p_noneok(value, 16,8)


class AllSensors (PackedAttr):

    def __init__(self, ColdWater=None, HotWater=None, MixedWater=None,
                 EspressoWater=None, InCaseAmbient=None, HeatSink=None,
                 SteamHeater=None, WaterHeater=None):
        super(AllSensors, self).__init__()
        self.ColdWater = ColdWater
        self.HotWater = HotWater
        self.MixedWater = MixedWater
        self.EspressoWater = EspressoWater
        self.InCaseAmbient = InCaseAmbient
        self.HeatSink = HeatSink
        self.SteamHeater = SteamHeater
        self.WaterHeater = WaterHeater

    def from_wire_bytes(self, wire_bytes, arrival_time=None):
        super(AllSensors, self).from_wire_bytes(wire_bytes, arrival_time)
        (
            self._ColdWater,
            self._HotWater,
            self._MixedWater,
            self._EspressoWater,
            self._InCaseAmbient,
            self._SteamHeater,
            self._WaterHeater,
         ) = unpack('>HHHHHHHH', wire_bytes)

        self._ColdWater = self._ColdWater / 2**8
        self._HotWater = self._HotWater / 2**8
        self._MixedWater = self._MixedWater / 2**8
        self._EspressoWater = self._EspressoWater / 2**8
        self._InCaseAmbient = self._InCaseAmbient / 2**8
        self._SteamHeater = self._SteamHeater / 2**8
        self._WaterHeater = self._WaterHeater / 2**8

        return self

    def as_wire_bytes(self):
        raise NotImplementedError

    def log_string(self):
        return 'Cold: {:.2f} Hot: {:.2f} Mix: {:.2f}' \
            'Espresso: {:.2f} Case: {:.2f} Steam: {:.2f} Water: {:.2f}'.format(
            self._ColdWater,
            self._HotWater,
            self._MixedWater,
            self._EspressoWater,
            self._InCaseAmbient,
            self._SteamHeater,
            self._WaterHeater,
        )

    @property
    def ColdWater(self):
        return self._ColdWater

    @ColdWater.setter
    def ColdWater(self, value):
        self._ColdWater = validate_u_p_noneok(value, 16, 8)

    @property
    def HotWater(self):
        return self._HotWater

    @HotWater.setter
    def HotWater(self, value):
        self._HotWater = validate_u_p_noneok(value, 16, 8)

    @property
    def MixedWater(self):
        return self._MixedWater

    @MixedWater.setter
    def MixedWater(self, value):
        self._MixedWater = validate_u_p_noneok(value, 16, 8)

    @property
    def EspressoWater(self):
        return self._EspressoWater

    @EspressoWater.setter
    def EspressoWater(self, value):
        self._EspressoWater = validate_u_p_noneok(value, 16, 8)

    @property
    def InCaseAmbient(self):
        return self._InCaseAmbient

    @InCaseAmbient.setter
    def InCaseAmbient(self, value):
        self._InCaseAmbient = validate_u_p_noneok(value, 16, 8)

    @property
    def HeatSink(self):
        return self._HeatSink

    @HeatSink.setter
    def HeatSink(self, value):
        self._HeatSink = validate_u_p_noneok(value, 16, 8)

    @property
    def SteamHeater(self):
        return self._SteamHeater

    @SteamHeater.setter
    def SteamHeater(self, value):
        self._SteamHeater = validate_u_p_noneok(value, 16, 8)

    @property
    def WaterHeater(self):
        return self._WaterHeater

    @WaterHeater.setter
    def WaterHeater(self, value):
        self._WaterHeater = validate_u_p_noneok(value, 16, 8)


# TODO; T_StoredShots, u1, NumberOfStoredShots (DE1 unimplemented?)


class TotalVOrWFlags (enum.IntFlag):
    UseVolume = 0x0000
    UseWeight = 0x8000


# TODO: TotalVOrW (DE1 unimplemented?)


class FrameFlags (enum.IntFlag):
    CtrlF       = 0x01  # Flow (or pressure)
    DoCompare   = 0x02  # Early exit if compare is true
    DC_GT       = 0x04  # Set for greater than coparison
    DC_CompF    = 0x08  # Compare flow (or pressure)
    TMixTemp    = 0x10  # Disable shower-head compensation (target MixTemp)
    Interpolate = 0x20  # Ramp to target (or jump)
    IgnoreLimit = 0x40  # Ignore minimum pressure and maximum flow settings

    DontInterpolate = 0x00
    CtrlP           = 0x00
    DC_CompP        = 0x00
    DC_LT           = 0x00
    TBasketTemp     = 0x00

    DontCompare     = 0x00
    ObserveLimit    = 0x00

    def not_flag_name(self):
        """
        Only works for bit-set flags, but only needed there
        """
        map = {
            FrameFlags.CtrlF:       'CtrlP',
            FrameFlags.DoCompare:   'DontCompare',
            FrameFlags.DC_GT:       'DC_LT',
            FrameFlags.DC_CompF:    'DC_CompP',
            FrameFlags.TMixTemp:    'TBasketTemp',
            FrameFlags.Interpolate: 'DontInterpolate',
            FrameFlags.IgnoreLimit: 'ObserveLimit',
        }
        return map.get(self, f"NOT{self.name}")




class ShotFrame (PackedAttr):

    def __init__(self, Flag=None, SetVal=None, Temp=None,
                 FrameLen=None, TriggerVal=None, MaxVol=None):
        super(ShotFrame, self).__init__()
        self.Flag = Flag
        self.SetVal = SetVal
        self.Temp = Temp
        self.FrameLen = FrameLen
        self.TriggerVal = TriggerVal
        self.MaxVol = MaxVol

    def from_wire_bytes(self, wire_bytes, arrival_time=None):
        super(ShotFrame, self).from_wire_bytes(wire_bytes, arrival_time)
        (self._Flag,
        self._SetVal,
        self._Temp,
        self._FrameLen,
        self._TriggerVal,
        self._MaxVol,
         ) = unpack('>BBBBBH', wire_bytes)

        self._SetVal = self._SetVal / 2**4
        self._Temp = self._Temp / 2
        self._FrameLen = f8_1_7_decode(self._FrameLen)
        self._TriggerVal = self._TriggerVal / 2**4

        return self

    def as_wire_bytes(self):
        return pack('>BBBBBH',
                    self._Flag,
                    p4(self._SetVal),
                    p1(self._Temp),
                    f8_1_7(self._FrameLen),
                    p4(self._TriggerVal),
                    p0(self._MaxVol),
                    )

    def log_string(self):
        flag_list = []
        for flag in FrameFlags:
            if flag.value and self._Flag & flag.value:
                flag_list.append(flag.name)
            elif flag.value:
                flag_list.append(flag.not_flag_name())
        if len(flag_list) > 1:
            flags = ','.join(flag_list)
        else:
            flags = '(none)'
        return '{} SetVal: {} Temp: {} Len: {} Trigger: {} MaxVol: {}'.format(
            flags,
            self._SetVal,
            self._Temp,
            self._FrameLen,
            self._TriggerVal,
            self._MaxVol,
        )

    @property
    def Flag(self):
        return self._Flag

    @Flag.setter
    def Flag(self, value):
        self._Flag = validate_u_p_noneok(value, 8, 0)

    @property
    def SetVal(self):
        return self._SetVal

    @SetVal.setter
    def SetVal(self, value):
        self._SetVal = validate_u_p_noneok(value, 8, 4)

    @property
    def Temp(self):
        return self._Temp

    @Temp.setter
    def Temp(self, value):
        self._Temp = validate_u_p_noneok(value, 8, 1)

    @property
    def FrameLen(self):
        return self._FrameLen

    @FrameLen.setter
    def FrameLen(self, value):
        self._FrameLen = validate_f8_1_7_noneok(value)

    @property
    def TriggerVal(self):
        return self._TriggerVal

    @TriggerVal.setter
    def TriggerVal(self, value):
        self._TriggerVal = validate_u_p_noneok(value, 8, 4)

    @property
    def MaxVol(self):
        return self._MaxVol

    @MaxVol.setter
    def MaxVol(self, value):
        self._MaxVol = validate_u_p_noneok(value, 10, 0)


class ShotExtFrame (PackedAttr):

    def __init__(self, MaxFlowOrPressure=None, MaxForPRange=None):
        super(ShotExtFrame, self).__init__()
        self.MaxFlowOrPressure = MaxFlowOrPressure
        self.MaxFoPRange = MaxForPRange

    def from_wire_bytes(self, wire_bytes, arrival_time=None):
        super(ShotExtFrame, self).from_wire_bytes(wire_bytes, arrival_time)
        (val, rng) = unpack('BBxxxxx', wire_bytes)
        self._MaxFlowOrPressure = val / 2**4
        self._MaxFoPRange = rng / 2**4

        return self

    def as_wire_bytes(self):
        return pack('BBxxxxx',
                    p4(self._MaxFlowOrPressure),
                    p4(self._MaxFoPRange)
                    )

    def log_string(self):
        return 'Limit: {:.2f} Range: {:.2f}'.format(
            self._MaxFlowOrPressure,
            self._MaxFoPRange,
        )

    @property
    def MaxFlowOrPressure(self):
        return self._MaxFlowOrPressure

    @MaxFlowOrPressure.setter
    def MaxFlowOrPressure(self, value):
        self._MaxFlowOrPressure = validate_u_p_noneok(value, 8, 4)

    @property
    def MaxFoPRange(self):
        return self._MaxFoPRange

    @MaxFoPRange.setter
    def MaxFoPRange(self, value):
        self._MaxFoPRange = validate_u_p_noneok(value, 8, 4)


class ShotTail (PackedAttr):

    def __init__(self, MaxTotalVolume=None, ignore_pi=True):
        super(ShotTail, self).__init__()
        self.MaxTotalVolume = MaxTotalVolume
        self.ignore_pi = ignore_pi

    def from_wire_bytes(self, wire_bytes, arrival_time=None):
        super(ShotTail, self).from_wire_bytes(wire_bytes, arrival_time)
        (value) = unpack('Hxxxxx', wire_bytes)
        self._MaxTotalVolume = value & 0x03ff
        self._ignore_pi = bool(value & 8000)

        return self

    def as_wire_bytes(self):
        value = p0(self._MaxTotalVolume)
        if not isinstance(self._ignore_pi, bool):
            raise DE1APIValueError("Expecting a bool for ignore_pi")
        if self._ignore_pi:
            value += 0x8000
        return pack('Hxxxxx', value)

    def log_string(self):
        return 'Limit: {} ignore_pi: {}'.format(
            self._MaxTotalVolume,
            self._ignore_pi,
        )

    @property
    def MaxTotalVolume(self):
        return self._MaxTotalVolume

    @MaxTotalVolume.setter
    def MaxTotalVolume(self, value):
        self._MaxTotalVolume = validate_u_p_noneok(
            value & 0x03ff, 10, 0
        )

    @property
    def ignore_pi(self):
        return self._ignore_pi

    @ignore_pi.setter
    def ignore_pi(self, value):
        if value is None:
            self._ignore_pi = None
        if isinstance(value, bool):
            self._ignore_pi = value
        if value == 0 or value == 1:
            self._ignore_pi = bool(value)
        else:
            raise DE1APIValueError("Expecting a bool for ignore_pi")


MAX_FRAMES = 20


class ShotDescHeader (PackedAttr):

    def __init__(self, HeaderV=None, NumberOfFrames=None,
                 NumberOfPreinfuseFrames=None,
                 MinimumPressure=None, MaximumFlow=None):
        super(ShotDescHeader, self).__init__()
        self.HeaderV = HeaderV
        # Declare internals directly so they can be cross-checked later
        self._NumberOfFrames = NumberOfFrames
        self._NumberOfPreinfuseFrames = NumberOfPreinfuseFrames
        self._check_frame_numbers()
        self._MinimumPressure = None
        self._MaximumFlow = None

        self.MinimumPressure = MinimumPressure
        self.MaximumFlow = MaximumFlow


    def from_wire_bytes(self, wire_bytes, arrival_time=None):
        super(ShotDescHeader, self).from_wire_bytes(wire_bytes, arrival_time)
        (
            self._HeaderV,
            self._NumberOfFrames,
            self._NumberOfPreinfuseFrames,
            self._MinimumPressure,
            self._MaximumFlow,
        ) = unpack('>BBBBB', wire_bytes)

        self._MinimumPressure = self._MinimumPressure / 2**4
        self._MaximumFlow = self._MaximumFlow / 2**4

        return self

    def as_wire_bytes(self):
        self._check_frame_numbers()
        return pack('>BBBBB',
                    p0(self._HeaderV),
                    p0(self._NumberOfFrames),
                    p0(self._NumberOfPreinfuseFrames),
                    p4(self._MinimumPressure),
                    p4(self._MaximumFlow),
                    )

    def log_string(self):
        return 'V{} Total: {} PI: {} MinPress {:.2f} MaxFlow {:.2f}'.format(
            self._HeaderV,
            self._NumberOfFrames,
            self._NumberOfPreinfuseFrames,
            self._MinimumPressure,
            self._MaximumFlow,
        )

    @property
    def HeaderV(self):
        return self._HeaderV

    @HeaderV.setter
    def HeaderV(self, value):
        if value != 1 and value is not None:  # "Set to 1 for this type of shot description"
            raise DE1APIValueError("HeaderV must be 1 at this time")
        self._HeaderV = value

    @property
    def NumberOfFrames(self):
        return self._NumberOfFrames

    @NumberOfFrames.setter
    def NumberOfFrames(self, value):
        self._check_frame_numbers(nf=value)
        self._NumberOfFrames = value

    @property
    def NumberOfPreinfuseFrames(self):
        return self._NumberOfPreinfuseFrames

    @NumberOfPreinfuseFrames.setter
    def NumberOfPreinfuseFrames(self, value):
        self._check_frame_numbers(npi=value)
        self._NumberOfPreinfuseFrames = value

    @property
    def MinimumPressure(self):
        return self._MinimumPressure

    @MinimumPressure.setter
    def MinimumPressure(self, value):
        self._MinimumPressure = validate_u_p_noneok(value, 8, 4)

    @property
    def MaximumFlow(self):
        return self._MaximumFlow

    @MaximumFlow.setter
    def MaximumFlow(self, value):
        self._MaximumFlow = validate_u_p_noneok(value, 8, 4)

    def _check_frame_numbers(self, npi=None, nf=None):
        if npi is None:
            npi = self._NumberOfPreinfuseFrames
        if nf is None:
            nf = self._NumberOfFrames
        if npi is not None:
            if npi > MAX_FRAMES:
                raise DE1APITooManyFramesError(f"{MAX_FRAMES}-frame limit")
            if nf is not None and npi > nf:
                raise DE1APIValueError(
                    "Number of preinfuse frames must not exceed number of frames")
        if nf is not None and nf > MAX_FRAMES:
            raise DE1APITooManyFramesError(f"{MAX_FRAMES}-frame limit")



# TODO: Deprecated (DE1 unimplemented?)
    """
    CUUID.Deprecated
    """

# class ShotDesc:
#   # Noted as what was on Deprecated before


class HeaderWrite (PackedAttr):

    cuuid = CUUID.HeaderWrite
    can_read = bool(cuuid is not None and cuuid.can_read)
    can_write = bool(cuuid is not None and cuuid.can_write)
    can_notify = bool(cuuid is not None and cuuid.can_notify)
    can_write_then_return = bool(cuuid is not None
                                 and cuuid.can_write_then_return)

    def __init__(self, Header=None):
        super(HeaderWrite, self).__init__()

        self.Header = Header

    def as_wire_bytes(self):
        return self._Header.as_wire_bytes()

    def from_wire_bytes(self, wire_bytes, arrival_time=None):
        super(HeaderWrite, self).from_wire_bytes(wire_bytes, arrival_time)
        self._Header.from_wire_bytes(wire_bytes, arrival_time)

        return self

    def log_string(self):
        return self._Header.log_string()

    @property
    def Header(self):
        return self._Header

    @Header.setter
    def Header(self, value):
        if value is None:
            value = ShotDescHeader()
        if not isinstance(value, ShotDescHeader):
            raise DE1APITypeError("Expected ShotDescHeader")
        self._Header = value


class FrameWrite (PackedAttr):

    cuuid = CUUID.FrameWrite
    can_read = bool(cuuid is not None and cuuid.can_read)
    can_write = bool(cuuid is not None and cuuid.can_write)
    can_notify = bool(cuuid is not None and cuuid.can_notify)
    can_write_then_return = bool(cuuid is not None
                                 and cuuid.can_write_then_return)

    def __init__(self, FrameToWrite: Optional[int] = None,
                 Frame: Optional[Union[ShotFrame,
                                       ShotTail,
                                       ShotExtFrame]] = None):
        super(FrameWrite, self).__init__()

        self._FrameToWrite = FrameToWrite
        self._Frame = Frame

    def as_wire_bytes(self):
        retval = bytearray(pack('>B', p0(self._FrameToWrite)))
        retval.extend(self._Frame.as_wire_bytes())
        return retval

    def from_wire_bytes(self, wire_bytes, arrival_time=None):
        raise NotImplementedError

    def log_string(self):
        return 'Frame #{} {}'.format(
            self._FrameToWrite,
            self._Frame.log_string()
        )

    @property
    def FrameToWrite(self):
        raise NotImplementedError

    @FrameToWrite.setter
    def FrameToWrite(self, value):
        raise NotImplementedError

    @property
    def Frame(self):
        raise NotImplementedError

    @Frame.setter
    def Frame(self, value):
        raise NotImplementedError


class FrameWrite_ShotFrame(FrameWrite):

    def __init__(self, FrameToWrite: Optional[int] = None,
                 Frame: Optional[ShotFrame] = None):
        super(FrameWrite_ShotFrame, self).__init__(FrameToWrite, Frame)


    def from_wire_bytes(self, wire_bytes, arrival_time=None):
        super(FrameWrite_ShotFrame, self).from_wire_bytes(wire_bytes,
                                                arrival_time)
        self.FrameToWrite = unpack('>B', wire_bytes[0:1])[0]
        self.Frame = ShotFrame().from_wire_bytes(wire_bytes[1:],
                                                 arrival_time)

        return self

    @property
    def FrameToWrite(self):
        return self._FrameToWrite

    @FrameToWrite.setter
    def FrameToWrite(self, value):
        if value is not None and not (
                (0 <= value < 20)
        ):
            raise DE1APIValueError(
                f"20-frame limit ({value})"
            )
        self._FrameToWrite = value

    @property
    def Frame(self):
        return self._Frame

    @Frame.setter
    def Frame(self, value):
        if not (isinstance(value, (ShotFrame,
                                   type(None)))):
            raise DE1APITypeError("Expecting ShotFrame")
        self._Frame = value


class FrameWrite_ShotExtFrame(FrameWrite):

    def __init__(self, FrameToWrite: Optional[int] = None,
                 Frame: Optional[ShotFrame] = None):
        super(FrameWrite_ShotExtFrame, self).__init__(FrameToWrite, Frame)

    def from_wire_bytes(self, wire_bytes, arrival_time=None):
        super(FrameWrite_ShotExtFrame, self).from_wire_bytes(wire_bytes,
                                                             arrival_time)
        self.FrameToWrite = unpack('>B', wire_bytes[0:1])[0]
        self.Frame = FrameWrite_ShotExtFrame().from_wire_bytes(wire_bytes[1:],
                                                               arrival_time)

        return self

    @property
    def FrameToWrite(self):
        return self._FrameToWrite

    @FrameToWrite.setter
    def FrameToWrite(self, value):
        if value is not None and not (
                (32 <= value < 52)
        ):
            raise DE1APIValueError(
                f"20-frame limit over 32-frame offset ({value})"
            )
        self._FrameToWrite = value

    @property
    def Frame(self):
        return self._Frame

    @Frame.setter
    def Frame(self, value):
        if not (isinstance(value, (ShotExtFrame,
                                   type(None)))):
            raise DE1APITypeError("Expecting ShotExtFrame")
        self._Frame = value



class FrameWrite_ShotTail(FrameWrite):

    def __init__(self, FrameToWrite: Optional[int] = None,
                 Frame: Optional[ShotTail] = None):
        super(FrameWrite_ShotTail, self).__init__(FrameToWrite, Frame)

    def from_wire_bytes(self, wire_bytes, arrival_time=None):
        super(FrameWrite_ShotTail, self).from_wire_bytes(wire_bytes,
                                                         arrival_time)
        self.FrameToWrite = unpack('>B', wire_bytes[0:1])[0]
        self.Frame = ShotTail().from_wire_bytes(wire_bytes[1:],
                                                arrival_time)

        return self

    @property
    def FrameToWrite(self):
        return self._FrameToWrite

    @FrameToWrite.setter
    def FrameToWrite(self, value):
        if value is not None and not (
                (0 < value <= 20)
        ):
            raise DE1APIValueError(
                f"20-frame limit ({value})"
            )
        self._FrameToWrite = value

    @property
    def Frame(self):
        return self._Frame

    @Frame.setter
    def Frame(self, value):
        if not (isinstance(value, (ShotTail, type(None)))):
            raise DE1APITypeError("Expecting ShotTail")
        self._Frame = value


class ShotState (PackedAttr):

    def __init__(self, GroupPressure=None, GroupFlow=None,
                 MixTemp=None, HeadTemp=None,
                 SetMixTemp=None, SetHeadTemp=None,
                 SetGroupPressure=None, SetGroupFlow=None,
                 FrameNumber=None, SteamTemp=None):
        super(ShotState, self).__init__()
        self.GroupPressure = GroupPressure
        self.GroupFlow = GroupFlow
        self.MixTemp = MixTemp
        self.HeadTemp = HeadTemp
        self.SetMixTemp = SetMixTemp
        self.SetHeadTemp = SetHeadTemp
        self.SetGroupPressure = SetGroupPressure
        self.SetGroupFlow = SetGroupFlow
        self.FrameNumber = FrameNumber
        self.SteamTemp = SteamTemp

    def from_wire_bytes(self, wire_bytes, arrival_time=None):
        super(ShotState, self).from_wire_bytes(wire_bytes, arrival_time)
        (
            gp, gf, mt, hth, htl, smt, sht, sgp, sgf, fn, st
        ) = unpack('>HHHBHHHBBBB', wire_bytes)
        self._GroupPressure = gp / 2 ** 12
        self._GroupFlow = gf / 2 ** 12
        self._MixTemp = mt / 2 ** 8
        self._HeadTemp = hth + (htl / 2 ** 16)
        self._SetMixTemp = smt / 2 ** 8
        self._SetHeadTemp = sht / 2 ** 8
        self._SetGroupPressure = sgp / 2 ** 4
        self._SetGroupFlow = sgf / 2 ** 4
        self._FrameNumber = fn
        self._SteamTemp = st

        return self

    def as_wire_bytes(self):
        raise NotImplementedError

    def log_string(self):
        return 'P: {:.2f} F: {:.2f} ' \
               'Mix: {:.2f} Head: {:.2f} ' \
               'SetMix: {:.2f} SetHead: {:.2f} ' \
               'SetP: {:.2f} SetF: {:.2f} ' \
               'Frame: {} Steam: {}'.format(
            self._GroupPressure,
            self._GroupFlow,
            self._MixTemp,
            self._HeadTemp,
            self._SetMixTemp,
            self._SetHeadTemp,
            self._SetGroupPressure,
            self._SetGroupFlow,
            self._FrameNumber,
            self._SteamTemp,
        )


    @property
    def GroupPressure(self):
        return self._GroupPressure

    @GroupPressure.setter
    def GroupPressure(self, value):
        self._GroupPressure = validate_u_p_noneok(value, 16, 12)

    @property
    def GroupFlow(self):
        return self._GroupFlow

    @GroupFlow.setter
    def GroupFlow(self, value):
        self._GroupFlow = validate_u_p_noneok(value, 16, 12)

    @property
    def MixTemp(self):
        return self._MixTemp

    @MixTemp.setter
    def MixTemp(self, value):
        self._MixTemp = validate_u_p_noneok(value, 16, 8)

    @property
    def HeadTemp(self):
        return self._HeadTemp

    @HeadTemp.setter
    def HeadTemp(self, value):
        self._HeadTemp = validate_u_p_noneok(value, 24, 16)

    @property
    def SetMixTemp(self):
        return self._SetMixTemp

    @SetMixTemp.setter
    def SetMixTemp(self, value):
        self._SetMixTemp = validate_u_p_noneok(value, 16, 8)

    @property
    def SetHeadTemp(self):
        return self._SetHeadTemp

    @SetHeadTemp.setter
    def SetHeadTemp(self, value):
        self._SetHeadTemp = validate_u_p_noneok(value, 16, 8)

    @property
    def SetGroupPressure(self):
        return self._SetGroupPressure

    @SetGroupPressure.setter
    def SetGroupPressure(self, value):
        self._SetGroupPressure = validate_u_p_noneok(value, 8, 4)

    @property
    def SetGroupFlow(self):
        return self._SetGroupFlow

    @SetGroupFlow.setter
    def SetGroupFlow(self, value):
        self._SetGroupFlow = validate_u_p_noneok(value, 8, 4)

    @property
    def FrameNumber(self):
        return self._FrameNumber

    @FrameNumber.setter
    def FrameNumber(self, value):
        self._FrameNumber = validate_u_p_noneok(value, 8, 0)

    @property
    def SteamTemp(self):
        return self._SteamTemp

    @SteamTemp.setter
    def SteamTemp(self, value):
        self._SteamTemp = validate_u_p_noneok(value, 8, 0)


class ShotSample (PackedAttr):

    cuuid = CUUID.ShotSample
    can_read = bool(cuuid is not None and cuuid.can_read)
    can_write = bool(cuuid is not None and cuuid.can_write)
    can_notify = bool(cuuid is not None and cuuid.can_notify)
    can_write_then_return = bool(cuuid is not None
                                 and cuuid.can_write_then_return)

    def __init__(self, SampleTime=None, State=None):
        super(ShotSample, self).__init__()

        self.SampleTime = SampleTime
        self.ShotState = State

    def from_wire_bytes(self, wire_bytes, arrival_time=None):
        super(ShotSample, self).from_wire_bytes(wire_bytes, arrival_time)
        self.SampleTime = unpack('>H', wire_bytes[0:2])[0]
        self.ShotState = ShotState().from_wire_bytes(wire_bytes[2:], arrival_time)

        return self

    def log_string(self):
        return 'Clock: {}: {}'.format(self._SampleTime, self._State.log_string())

    @property
    def SampleTime(self):
        return self._SampleTime

    @SampleTime.setter
    def SampleTime(self, value):
        self._SampleTime = validate_u_p_noneok(value, 16, 0)

    @property
    def ShotState(self):
        return self._State

    @ShotState.setter
    def ShotState(self, value):
        if not (isinstance(value, ShotState) or value is None):
            raise DE1APITypeError("Expected ShotState")
        if value is None:
            value = ShotState()
        self._State = value

    #
    # Convenience methods to go straight to the State
    #

    @property
    def GroupPressure(self):
        return self._State.GroupPressure

    @property
    def GroupFlow(self):
        return self._State.GroupFlow

    @property
    def MixTemp(self):
        return self._State.MixTemp

    @property
    def HeadTemp(self):
        return self._State.HeadTemp

    @property
    def SetMixTemp(self):
        return self._State.SetMixTemp

    @property
    def SetHeadTemp(self):
        return self._State.SetHeadTemp

    @property
    def SetGroupPressure(self):
        return self._State.SetGroupPressure

    @property
    def SetGroupFlow(self):
        return self._State.SetGroupFlow

    @property
    def FrameNumber(self):
        return self._State.FrameNumber

    @property
    def SteamTemp(self):
        return self._State.SteamTemp


# TODO: ShotData (DE1 unimplemented?)


# TODO: ShotDirectory (DE1 unimplemented?)
    """
    CUUID.ShotDirectory
    """

class FWImageInfo (PackedAttr):

    def __init__(self, Version=None, Hash=None):
        super(FWImageInfo, self).__init__()
        self.Version = Version
        self.Hash = Hash

    def from_wire_bytes(self, wire_bytes, arrival_time=None):
        super(FWImageInfo, self).from_wire_bytes(wire_bytes, arrival_time)
        (
            self._Version,
            self._Hash,
        ) = unpack('>IQ', wire_bytes)

        return self

    def as_wire_bytes(self):
        return pack('>IQ',
                    u(self._Version),
                    u(self._Hash),
                    )

    def log_string(self):
        return 'Version: {} Hash: {}'.format(self._Version, self._Hash)

    @property
    def Version(self):
        return self._Version

    @Version.setter
    def Version(self, value):
        self._Version = validate_u_p_noneok(value, 32, 0)

    @property
    def Hash(self):
        return self._Hash

    @Hash.setter
    def Hash(self, value):
        self._Hash = validate_u_p_noneok(value, 64, 0)


# TODO: FirmwareImages (DE1 unimplemented?)


class MoveMMRWindow (PackedAttr):

    def __init__(self, Offset=None, Len=None):
        super(MoveMMRWindow, self).__init__()
        self.Offset = Offset
        self.Len = Len

    def from_wire_bytes(self, wire_bytes, arrival_time=None):
        super(MoveMMRWindow, self).from_wire_bytes(wire_bytes, arrival_time)
        (
            self._Offset,
            self._Len
        ) = unpack('IB', wire_bytes)

        return self

    def as_wire_bytes(self):
        return pack('IB', u(self._Len), p0(self._Offset))

    def log_string(self):
        return 'Offset: {} Len: {}'.format(self._Offset, self._Len)

    @property
    def Offset(self):
        return self._Offset

    @Offset.setter
    def Offset(self, value):
        self._Offset = validate_u_p_noneok(value, 32, 0)

    @property
    def Len(self):
        return self._Len

    @Len.setter
    def Len(self, value):
        self._Len = validate_u_p_noneok(value, 64, 0)


class MMRData (PackedAttr):
    """
    This represents generic, MMR data
    """

    def __init__(self, Len=None, Address=None,
                 addr_high=None, addr_low=None,
                 Data=None):
        """
        Generic MMR struct, either for read or write

        :param Len:         NB: Len is (words + 1) so 1-256 words, 4-1024 bytes
        :param Address:     Either a full, 3-byte address, or
        :param addr_high:   High byte of address and
        :param addr_low:    Low, two bytes of address
        :param Data:        Byte data
        """
        super(MMRData, self).__init__()
        if Address is not None and (addr_high is not None or addr_low is not None):
            raise MMRTypeError("Either Address or addr_high and addr_low can be specified")
        self.Len: Union[int, None] = Len
        self.addr_low: Union[int, None] = addr_low
        self.addr_high: Union[int, None] = addr_high
        self.Data: Union[bytes, bytearray, None] = Data
        self.Address: Union[int, None] = Address

    def from_wire_bytes(self, wire_bytes, arrival_time=None):
        super(MMRData, self).from_wire_bytes(wire_bytes, arrival_time)
        (self._Len, self._addr_high, self._addr_low) \
            = unpack('>BBH', wire_bytes[0:4])
        """
        Read request interprets Len as (words - 1) requested
        This at least won't error-out
        """
        self._Data = wire_bytes[4:(self._Len + 4)]

        return self

    # TODO: Is padding to 16 bytes a hack or not?

    def as_wire_bytes(self):
        retval = bytearray(pack('>BBH',
                                p0(self._Len),
                                int(self._addr_high),
                                int(self._addr_low))
                           )
        retval.extend(self._padded_data())
        return retval

    def log_string(self):
        if self._addr_high == 0x80:
            addr_name = MMR0x80LowAddr.for_logging(self.addr_low)
        else:
            addr_name = ""
        if len(addr_name) > 0:
            addr_name = " " + addr_name
        return '0x{:02x} ({}): 0x{:02x} {:04x}{}: {} ({})'.format(
            self._Len,
            self._Len,
            self._addr_high,
            self._addr_low,
            addr_name,
            data_as_hex(self._Data),
            len(self._Data),
        )

    @property
    def Len(self):
        return self._Len

    @Len.setter
    def Len(self, value):
        self._Len = validate_u_p_noneok(value, 8, 0)

    @property
    def Address(self):
        if self._addr_high and self.addr_low:
            return (self._addr_high << 16) + self._addr_low
        else:
            return None

    @Address.setter
    def Address(self, value):
        addr = validate_u_p_noneok(value, 24, 0)
        if addr is not None:
            self.addr_high = (addr & 0xff0000) >> 16
            self.addr_low = addr & 0x00ffff

    @property
    def addr_high(self):
        return self._addr_high

    @addr_high.setter
    def addr_high(self, value):
        self._addr_high = validate_u_p_noneok(value, 8, 0)

    @property
    def addr_low(self):
        return self._addr_low

    @addr_low.setter
    def addr_low(self, value):
        self._addr_low = validate_u_p_noneok(value, 16, 0)

    @property
    def Data(self):
        return self._Data

    @Data.setter
    def Data(self, value):
        if value is not None:
            if not isinstance(value, (bytes, bytearray)):
                raise MMRTypeError("Expected bytes or bytearray")
            if len(value) > 16:
                raise MMRTypeError("MMRData is limited to 16 bytes")
        self._Data = value

    def _padded_data(self):
        if self._Data is None:
            retval = bytearray([0] * 16)
        else:
            retval = bytearray(self._Data)
            if len(retval) < 16:
                retval.extend([0] * (16 - len(retval)))
            if len(retval) > 16:
                raise MMRDataTooLongError("MMRData is limited to 16 bytes")
        return retval


class ReadFromMMR (MMRData):
    """
    An MMR read request interprets the Len field as (words + 1) requested
    A read response interprets it as valid bytes in the payload

    NB: Default for from_response is
        False for ReadFromMMR() -- likely an outgoing request
        True for ReadFromMMR().from_bytes() -- likely an incoming response
            though will not override an already-set value
    """

    cuuid = CUUID.ReadFromMMR
    can_read = bool(cuuid is not None and cuuid.can_read)
    can_write = bool(cuuid is not None and cuuid.can_write)
    can_notify = bool(cuuid is not None and cuuid.can_notify)
    can_write_then_return = bool(cuuid is not None
                                 and cuuid.can_write_then_return)

    def __init__(self, Len=None, Address=None,
                 addr_high=None, addr_low=None,
                 Data=None, from_response=False):

        # TODO: Remove this hack:
        if Len is not None and not isinstance(Len, int):
            tb = "".join(traceback.format_stack(limit=3))
            intLen = int(Len)
            if intLen != Len:
                logger.error(f"Len {Len} is not an int:\n{tb}")
            else:
                logger.info(f"Len {Len} is not an int:\n{tb}")
            Len = intLen

        super(ReadFromMMR, self).__init__(Len=Len, Address=Address,
                    addr_high=addr_high, addr_low=addr_low, Data=Data)

        self._from_response = from_response
        if Len is not None:
            if not self._from_response:
                if not 0 <= Len <= 255:
                    raise MMRValueError(
                        "MMR read requests are limited to 255 + 1 words"
                    )
            elif Len is not None:
                if not 0 <= Len < 16:
                    raise MMRValueError(
                        "MMR read responses are limited to 16 bytes"
                    )


    def from_wire_bytes(self, wire_bytes, arrival_time=None,
                        from_response=True):
        super(ReadFromMMR, self).from_wire_bytes(wire_bytes, arrival_time)
        if from_response:
            self._Data = wire_bytes[4:(self._Len + 4)]
        else:
            # No information on how long, so take what's there
            # super() only takes Len bytes, which may be 0
            # on a one-word read request
            self._Data = wire_bytes[4:]

        return self

    @property
    def is_within_debug_log(self):
        if self.Address is None:
            return False
        if self._from_response:
            end = self.addr_low + self.Len - 1
        else:
            end = self.addr_low + 4 * self.Len - 1
        return (
                (MMR0x80LowAddr.DEBUG_BUFFER
                 <= self.addr_low < MMR0x80LowAddr.DEBUG_CONFIG)
                and (MMR0x80LowAddr.DEBUG_BUFFER
                     <= end < MMR0x80LowAddr.DEBUG_CONFIG)
                and self.addr_high == 0x80
        )


class WriteToMMR (MMRData):

    cuuid = CUUID.WriteToMMR
    can_read = bool(cuuid is not None and cuuid.can_read)
    can_write = bool(cuuid is not None and cuuid.can_write)
    can_notify = bool(cuuid is not None and cuuid.can_notify)
    can_write_then_return = bool(cuuid is not None
                                 and cuuid.can_write_then_return)

    def __init__(self, Len=None, Address=None,
                 addr_high=None, addr_low=None,
                 Data=None):

        if Len is None and Data is not None:
            Len = len(Data)

        super(WriteToMMR, self).__init__(Len=Len, Address=Address,
                 addr_high=addr_high, addr_low=addr_low, Data=Data)

        self._check_data_len_consistent()


    # Len and Data consistent at creation
    # Len consistent if Data not None
    # Data consistent if Len is not None

    def _check_data_len_consistent(self, d=None, given_len=None):
        if d is None:
            d = self._Data
        if given_len is None:
            given_len = self._Len

        if (given_len is not None) and (given_len > 16):
            raise MMRValueError(
                f"MMR writes are limited to 16 bytes ({given_len} requested)"
            )

        if (d is not None) and (len(d) > 16):
            raise MMRValueError(
                "MMR writes are limited to 16 bytes (len(d) data length)"
            )

        if (d is not None) and (given_len is not None) \
                and (len(d) != given_len):
            raise MMRValueError(
                f"MMR write Len of {given_len} is not length of data ({len(d)}"
            )


# TODO: ShotMapRequest (DE1 unimplemented?)
    """
    CUUID.ShotMapRequest
    """


class FWErrorMapRequest(enum.IntEnum):
    Ignore = 0x0
    ReportFirst = 0xffffff
    ReportNext = 0xfffffe

class FWErrorMapResponse(enum.IntEnum):
    NoneFound = 0xfffffd

class FWMapRequest(PackedAttr):
    """
    See also FWMapRequestResponse

    // Request that a firmware image be put in the memory mapped region for later reading
    U16P0 WindowIncrement;  // Every time the MMR is read, add this byte offset to the MMR base address
    U8P0  FWToErase;        // If this field is non-zero, erase the firmware slot in question. (1 or 2)
                            // Stays non-zero until firmware is erased
    U8P0  FWToMap;          // Either 1 or 2

    // How to use FirstError:
    // Enable Notify for this characteristic
    // Write 0xFFFFFF to FirstError, FWToMap should be set to the required image. FWErase should not be set.
    // A notify will later arrive with the first address that needs repairing.
    // Write 0xFFFFFE in order to update FirstError to the next block that needs repairing.
    // 0xFFFFFF resets the search address to 0, 0xFFFFFE does not.
    // If there are no remaining errors, the notified FirstError value will be 0xFFFFFD
    U24P0 FirstError;
    """

    cuuid = CUUID.FWMapRequest
    can_read = bool(cuuid is not None and cuuid.can_read)
    can_write = bool(cuuid is not None and cuuid.can_write)
    can_notify = bool(cuuid is not None and cuuid.can_notify)
    can_write_then_return = bool(cuuid is not None
                                 and cuuid.can_write_then_return)

    # TODO: FirstError is ambiguous as 0x00 0000 is either
    #       * Response of Ignore from an erase request
    #       * An error at address 0x0 from an error search

    def __init__(self, WindowIncrement=None, FWToErase=None, FWToMap=None,
                 FirstError=None, arrival_time=None, from_response=False):
        super(FWMapRequest).__init__()

        self._WindowIncrement: Optional[int] = None
        self._FWToErase: Optional[int] = None
        self._FWToMap: Optional[int] = None
        self._FirstError: Optional[int] = None
        self._arrival_time: Optional[float] = arrival_time
        self._from_response: bool = from_response

        self.WindowIncrement = WindowIncrement
        self.FWToErase = FWToErase
        self.FWToMap = FWToMap
        self.FirstError = FirstError

    @property
    def WindowIncrement(self):
        return self._WindowIncrement

    @WindowIncrement.setter
    def WindowIncrement(self, value):
        if not isinstance(value, (int, type(None))):
            raise DE1APITypeError(
                f"WindowIncrement needs to be a non-negative integer: {value}")
        self._WindowIncrement = validate_u_p_noneok(value, 16, 0)

    def _check_inconsistent_fw_values(self):
        if (self.FWToErase in (1, 2)
                and self.FWToMap is not None
                and self.FWToErase != self.FWToMap):
            logger.warning(
                "FWToErase inconsistent with FWToMap: "
                f"{self.FWToErase} != {self.FWToMap}")

    @property
    def FWToErase(self):
        return self._FWToErase

    @FWToErase.setter
    def FWToErase(self, value):
        if value not in (0, 1, 2, None):
            raise DE1APIValueError(f"FWToErase must be 0, 1, or 2: {value}")
        if value is not None:
            value = int(value)
            if (value > 0
                    and self.FirstError is not None
                    and self.FirstError != FWErrorMapRequest.Ignore):
                raise DE1APIValueError(
                    "Erasing firmware while checking for upload errors "
                    f"to FW {value} with {self.FirstError.__repr__()}"
                )
        self._FWToErase = value
        self._check_inconsistent_fw_values()

    @property
    def FWToMap(self):
        return self._FWToMap

    @FWToMap.setter
    def FWToMap(self, value):
        if value not in (1, 2, None):
            raise DE1APIValueError(f"FWToMap must be 1 or 2: {value}")
        if value is not None:
            value = int(value)
        self._FWToMap = value
        self._check_inconsistent_fw_values()

    @property
    def FirstError(self):
        return self._FirstError

    @FirstError.setter
    def FirstError(self, value):
        if value is not None:

            if self._from_response:
                try:
                    value = FWErrorMapResponse(value)
                except ValueError:
                    pass

            else:
                try:
                    value = FWErrorMapRequest(value)
                except ValueError as e:
                    raise DE1APIValueError(e)
                if self.FWToErase > 0 and value != FWErrorMapRequest.Ignore:
                    raise DE1APIValueError(
                        "Erasing firmware while checking for upload errors: "
                        f"to FW {self.FWToErase} with {value.__repr__()}"
                    )
        self._FirstError = value

    def as_wire_bytes(self) -> Union[bytes, bytearray]:
        return pack('>HBBBH',
                    self.WindowIncrement,
                    self.FWToErase,
                    self.FWToMap,
                    self.FirstError >> 16,
                    self.FirstError & 0xffff
                    )

    def from_wire_bytes(self, wire_bytes: Union[bytes, bytearray],
                        arrival_time=None, from_response=True):
        self._from_response = from_response
        (self.WindowIncrement,
         self.FWToErase,
         self.FWToMap,
         fe_high,
         fe_low,) = unpack('>HBBBH', wire_bytes)
        self.FirstError = (fe_high << 16) + fe_low
        self._arrival_time = arrival_time
        return self

    def log_string(self):

        # TODO: Make sure the other log strings are None-safe

        if self.WindowIncrement is not None:
            wi = f"0x{self.WindowIncrement:x}"
        else:
            wi = None
        if isinstance(self.FirstError, (FWErrorMapRequest, FWErrorMapResponse)):
            fe_str = self.FirstError.__str__()
        else:
            fe_high = self.FirstError >> 16
            fe_low = self.FirstError & 0xffff
            fe_str = f"Next error: 0x{fe_high:02x} {fe_low:04x}"

        return f"WinInc: {wi} Erase: {self.FWToErase} Map: {self.FWToMap} {fe_str}"



# TODO: DeleteShotRange (DE1 unimplemented?)
    """
    CUUID.DeleteShotRange
    """

class SetTime (PackedAttr):

    cuuid = CUUID.SetTime
    can_read = bool(cuuid is not None and cuuid.can_read)
    can_write = bool(cuuid is not None and cuuid.can_write)
    can_notify = bool(cuuid is not None and cuuid.can_notify)
    can_write_then_return = bool(cuuid is not None
                                 and cuuid.can_write_then_return)

    def __init__(self, Timestamp=None):
        super(SetTime, self).__init__()

        self.Timestamp = None

    def from_wire_bytes(self, wire_bytes, arrival_time=None):
        super(SetTime, self).from_wire_bytes(wire_bytes, arrival_time)
        self.Timestamp = unpack('>Q', wire_bytes)[0]

        return self

    def as_wire_bytes(self):
        return pack('>Q', u(self.Timestamp))

    def log_string(self):
        return '{}: {}'.format(
            self.Timestamp,
            time.ctime(self.Timestamp),
        )

    @property
    def Timestamp(self):
        return self._Timestamp

    @Timestamp.setter
    def Timestamp(self, value):
        self._Timestamp = validate_u_p_noneok(value, 64, 0)

    def from_seconds(self, seconds):
        self.Timestamp = int(round(seconds))

        return self



# TODO: TestReq (DE1 unimplemented?)


# TODO: CalTargets
# TODO: CalCommand
# TODO: Calibration


"""
CUUID.Calibration
"""

class CalTargets (enum.IntEnum):
    CalFlow     = 0     # ratiometric
    CalPressure = 1     # ratiometric
    CalTemp     = 2     # differential, degrees C
    CalError    = 255   # return value indicating bad request


class CalCommand (enum.IntEnum):
    Read        = 0
    Write       = 1
    Reset       = 2     # to factory value
    ReadFactory = 3


class Calibration (PackedAttr):
    """
    Used to read and write internal DE1 calibration parameters

    The WriteKey is None, by default, which is not valid to write
    It is available for configuration through the config instance
    """

    cuuid = CUUID.Calibration
    can_read = bool(cuuid is not None and cuuid.can_read)
    can_write = bool(cuuid is not None and cuuid.can_write)
    can_notify = bool(cuuid is not None and cuuid.can_notify)
    can_write_then_return = bool(cuuid is not None
                                 and cuuid.can_write_then_return)

    def __init__(self, WriteKey = 0,
                 CalCommand: Optional[CalCommand] = None,
                 CalTarget: Optional[CalTargets] = None,
                 DE1ReportedValue = 0,
                 MeasuredVal = 0):

        super(Calibration, self).__init__()

        self.WriteKey = WriteKey
        self._cal_command = None
        self._cal_target = None
        self._reported = 0
        self._measured = 0

        self.CalCommand = CalCommand
        self.CalTarget = CalTarget
        self.DE1ReportedVal = DE1ReportedValue
        self.MeasuredVal = MeasuredVal

    @property
    def CalCommand(self):
        return self._cal_command

    @CalCommand.setter
    def CalCommand(self, value):
        if isinstance(value, CalCommand) or value is None:
            self._cal_command = value
        else:
            raise DE1APITypeError(
                f"Expected CalCommand, not {type(value)}"
            )

    @property
    def CalTarget(self):
        return self._cal_target

    @CalTarget.setter
    def CalTarget(self, value):
        if isinstance(value, CalTargets) or value is None:
            self._cal_target = value
        else:
            raise DE1APITypeError(
                f"Expected CalTarget, not {type(value)}"
            )

    @property
    def DE1ReportedVal(self):
        return self._reported

    @DE1ReportedVal.setter
    def DE1ReportedVal(self, value):
        self._reported = validate_s_p(value, 32, 16)

    @property
    def MeasuredVal(self):
        return self._measured

    @MeasuredVal.setter
    def MeasuredVal(self, value):
        self._measured = validate_s_p(value, 32, 16)

    @property
    def is_error_response(self):
        return self.CalTarget == CalTargets.CalError

    @property
    def is_offset(self):
        return self == CalTargets.CalTemp

    @property
    def is_ratio(self):
        return not self.is_offset

    def from_wire_bytes(self, wire_bytes: Union[bytes, bytearray],
                        arrival_time=None):
        super(Calibration, self).from_wire_bytes(wire_bytes, arrival_time)
        (wk, cc, ct, rv, mv) = unpack('>IBBii', wire_bytes)
        self.WriteKey = wk
        self.CalCommand = CalCommand(cc)
        self.CalTarget = CalTargets(ct)
        self.DE1ReportedVal = rv / 2**16
        self.MeasuredVal = mv / 2**16

        return self

    def as_wire_bytes(self) -> Union[bytes, bytearray]:
        if self.CalTarget is None:
            raise DE1APIValueError("No CalTarget supplied")
        if self.CalCommand is None:
            raise DE1APIValueError("No CalCommand supplied")
        if self.CalCommand == CalCommand.Write and not self.WriteKey:
            raise DE1APIValueError(
                "WriteKey needed to write Calibration data")
        return pack('>IBBii',
                    int(self.WriteKey),
                    self.CalCommand.value,
                    self.CalTarget.value,
                    p16(self.DE1ReportedVal),
                    p16(self.MeasuredVal))

    @staticmethod
    def _val_str(val):
        if int(val) == val:
            return str(val)
        else:
            # 2^16 ~ 1.5e-5
            return f"{val:.5f}"

    def log_string(self):
        if self.is_error_response:
            erf = 'ERROR: '
        else:
            erf = ''
        return "{}{} {} de1: {} meas: {}".format(
            erf,
            self.CalCommand.name,
            self.CalTarget.name,
            self._val_str(self.DE1ReportedVal),
            self._val_str(self.MeasuredVal)
        )


# End of APIDataTypes.hpp


# From comments in APIDataTypes.hpp


class MMRV13ModelCode (enum.IntEnum):
    UNSET       =  0
    DE1         =  1
    DE1PLUS     =  2
    DE1PRO      =  3
    DE1XL       =  4
    DE1CAFE     =  5


class MMRGHCInfoBitMask (enum.IntFlag):
    NONE_SET                    = 0x00
    LED_CONTROLLER_PRESENT      = 0x01
    TOUCH_CONTROLLER_PRESENT    = 0x02
    GHC_ACTIVE                  = 0x04
    FACTORY_MODE          = 0x80000000


class MMR0x80LowAddr (enum.IntEnum):
    HW_CONFIG               = 0x0000
    MODEL                   = 0x0004
    CPU_BOARD_MODEL         = 0x0008
    V13_MODEL               = 0x000c
    CPU_FIRMWARE_BUILD      = 0x0010
    DEBUG_LEN               = 0x2800
    DEBUG_BUFFER            = 0x2804
    DEBUG_CONFIG            = 0x3804
    FAN_THRESHOLD           = 0x3808
    TANK_TEMP               = 0x380c
    HEATER_UP1_FLOW         = 0x3810
    HEATER_UP2_FLOW         = 0x3814
    WATER_HEATER_IDLE_TEMP  = 0x3818
    GHC_INFO                = 0x381c
    PREF_GHC_MCI            = 0x3820
    MAX_SHOT_PRESS          = 0x3824
    TARGET_STEAM_FLOW       = 0x3828
    STEAM_START_SECS        = 0x382c
    SERIAL_NUMBER           = 0x3830
    HEATER_VOLTAGE          = 0x3834    # NB: +1000 if manually set
    HEATER_UP2_TIMEOUT      = 0x3838
    CAL_FLOW_EST            = 0x383c
    FLUSH_FLOW_RATE         = 0x3840    # 60 default
    FLUSH_TEMP              = 0x3844    # 850 default
    FLUSH_TIMEOUT           = 0x3848    # 200 default
    HOT_WATER_FLOW_RATE     = 0x384c

    LAST_KNOWN              = 0x384c        # See also FeatureFlag

    #
    # Surprisingly, properties can be added with member data
    # even with an IntEnum
    #

    def __init__(self, addr):
        # super(MMR0x80LowAddr, self).__init__(addr)
        # enum.IntEnum.__init__(addr)
        self.last_requested: Optional[float] = None
        self.last_updated: Optional[float] = None
        self._data_ready_event: asyncio.Event = asyncio.Event()
        self._data_raw: Optional[Union[bytes, bytearray]] = None
        self._data_decoded: Optional[Union[bytes, bytearray,
                                           str, int, float, bool,
                                           dict]] = None

    @property
    def can_read(self):
        return self not in {
            MMR0x80LowAddr.PREF_GHC_MCI,
            MMR0x80LowAddr.MAX_SHOT_PRESS,
        }

    @property
    def can_write(self):
        return self in {
            MMR0x80LowAddr.FAN_THRESHOLD,
            MMR0x80LowAddr.TANK_TEMP,
            MMR0x80LowAddr.HEATER_UP1_FLOW,
            MMR0x80LowAddr.HEATER_UP2_FLOW,
            MMR0x80LowAddr.WATER_HEATER_IDLE_TEMP,
            MMR0x80LowAddr.TARGET_STEAM_FLOW,
            MMR0x80LowAddr.STEAM_START_SECS,
            MMR0x80LowAddr.HEATER_VOLTAGE,   # Needed for older machines
            MMR0x80LowAddr.HEATER_UP2_TIMEOUT,
            MMR0x80LowAddr.CAL_FLOW_EST,
            MMR0x80LowAddr.FLUSH_FLOW_RATE,
            MMR0x80LowAddr.FLUSH_TEMP,
            MMR0x80LowAddr.FLUSH_TIMEOUT,
            MMR0x80LowAddr.HOT_WATER_FLOW_RATE,
        }

    @classmethod
    def in_debug_buffer(cls, addr_low):
        return cls.DEBUG_BUFFER.value <= addr_low < cls.DEBUG_CONFIG.value

    @property
    def read_once(self) -> bool:
        return self.can_read and not self.can_write \
            and not MMR0x80LowAddr.in_debug_buffer(self.value) \
            and self != MMR0x80LowAddr.CPU_FIRMWARE_BUILD

    @classmethod
    def for_logging(cls, addr_low):
        try:
            addr_name = cls(addr_low).name.title()
            # The above catches the at-start condition
        except ValueError:
            if cls.in_debug_buffer(addr_low):
                addr_name = "{}+0x{:03x}".format(
                    cls.DEBUG_BUFFER.name.title(),
                    addr_low - cls.DEBUG_BUFFER.value,
                )
            else:
                addr_name = ""
        return addr_name

    def __repr__(self):
        return "<%s.%s: 0x%04x>" % (
                self.__class__.__name__, self._name_, self._value_)


def decode_one_mmr(addr_high: int, addr_low: Union[MMR0x80LowAddr, int],
                   mmr_bytes: Union[bytes, bytearray]):

    if addr_high != 0x80:
        # Unknown how to decode
        retval = mmr_bytes

    if addr_low in {
        MMR0x80LowAddr.HW_CONFIG,
        MMR0x80LowAddr.MODEL,
        MMR0x80LowAddr.SERIAL_NUMBER,
        MMR0x80LowAddr.DEBUG_CONFIG,
    }:
        # Unknown how to decode
        retval = mmr_bytes

    elif addr_low in {
        MMR0x80LowAddr.CPU_BOARD_MODEL,
        MMR0x80LowAddr.CAL_FLOW_EST,
    }:
        val = unpack('<I', mmr_bytes)[0]
        retval = val / 1000

    elif addr_low in {
        MMR0x80LowAddr.TARGET_STEAM_FLOW,
        MMR0x80LowAddr.STEAM_START_SECS,
    }:
        val = unpack('<I', mmr_bytes)[0]
        retval = val / 100

    elif addr_low in {
        MMR0x80LowAddr.HEATER_UP1_FLOW,
        MMR0x80LowAddr.HEATER_UP2_FLOW,
        MMR0x80LowAddr.WATER_HEATER_IDLE_TEMP,
        MMR0x80LowAddr.HEATER_UP2_TIMEOUT,
        MMR0x80LowAddr.FLUSH_FLOW_RATE,
        MMR0x80LowAddr.FLUSH_TEMP,
        MMR0x80LowAddr.FLUSH_TIMEOUT,
        MMR0x80LowAddr.HOT_WATER_FLOW_RATE,
    }:
        val = unpack('<I', mmr_bytes)[0]
        retval = val / 10

    elif MMR0x80LowAddr.in_debug_buffer(addr_low):
        retval = mmr_bytes.decode('utf-8')

    elif addr_low == MMR0x80LowAddr.V13_MODEL:
        retval = MMRV13ModelCode(unpack('<I', mmr_bytes)[0])

    elif addr_low == MMR0x80LowAddr.GHC_INFO:
        val = unpack('<I', mmr_bytes)[0]
        # retval = {}
        # for bit in MMRGHCInfoBitMask:
        #     retval[bit] = val & bit > 0
        retval = MMRGHCInfoBitMask(val)

    elif addr_low in {
        MMR0x80LowAddr.PREF_GHC_MCI,
        MMR0x80LowAddr.MAX_SHOT_PRESS,
    } or addr_low > MMR0x80LowAddr.LAST_KNOWN:
        # These aren't implemented (fully)
        retval = unpack('<I', mmr_bytes)[0]
        logger.warning(
            "Unexpected decode requested for "
            f"0x{addr_low:04x} {retval} 0x{retval:x} "
            "(unimplemented MMR)"
        )

    else:
        retval = unpack('<I', mmr_bytes)[0]

    return retval


def pack_one_mmr0x80_write(addr_low: MMR0x80LowAddr,
                             value: Union[float, int]) -> WriteToMMR:

    if not addr_low.can_write:
        raise DE1APIValueError(
            f"Not encoding a non-writable MMR target address: {addr_low}")

    if addr_low == MMR0x80LowAddr.CAL_FLOW_EST:
        binval = pack('<I', int(round(value * 1000)))

    elif addr_low == MMR0x80LowAddr.TARGET_STEAM_FLOW:
        # Check for 0.6 to 0.8
        binval = pack('<I', int(round(value * 100)))

    elif addr_low == MMR0x80LowAddr.STEAM_START_SECS:
        # Check for 0 to 4
        binval = pack('<I', int(round(value * 100)))

    elif addr_low in (
        MMR0x80LowAddr.HEATER_UP1_FLOW,
        MMR0x80LowAddr.HEATER_UP2_FLOW,
        MMR0x80LowAddr.WATER_HEATER_IDLE_TEMP,
        MMR0x80LowAddr.HEATER_UP2_TIMEOUT,
        MMR0x80LowAddr.FLUSH_FLOW_RATE,
        MMR0x80LowAddr.FLUSH_TEMP,
        MMR0x80LowAddr.FLUSH_TIMEOUT,
    ):
        binval = pack('<I', int(round(value * 10)))

    elif addr_low in (
            MMR0x80LowAddr.TANK_TEMP,  # 0-60°C ?
            MMR0x80LowAddr.FAN_THRESHOLD,
    ):
        binval = pack('<I', int(round(value)))

    else:
        raise DE1APIValueError(
            f"Not encoding an unrecognized MMR target address: {addr_low}")

    if addr_low == MMR0x80LowAddr.FAN_THRESHOLD:
        if not (0 <= value <= 60):
            logger.warning(
                f"Fan threshold seems out of range: 0 <= {value} <= 60")

    return WriteToMMR(
        addr_high=0x80,
        addr_low=addr_low,
        Data=binval,
    )



# Define the reverse mapping based on the subclasses defined
_cuuid_to_packed_attr_class = {}
for cpa in PackedAttr.__subclasses__():
    if (cuuid := cpa.cuuid) is not None:
        _cuuid_to_packed_attr_class[cuuid] = cpa


def packed_attr_from_cuuid(cuuid: CUUID,
                           wire_bytes: Optional[Union[bytes, bytearray]] = None
                           ) -> PackedAttr:
    try:
        pa = _cuuid_to_packed_attr_class[cuuid]()
        if wire_bytes is not None:
            pa = pa.from_wire_bytes(wire_bytes)
    except KeyError:
        pa = None
    return pa




"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only

The BlueZ stack will keep a connection to the device open even if Python exits.
This wrapper "automatically" adds an atexit close of the connection, as well as
trying to remove that atexit command if the object's __del__ method gets called.

There is only a handful of these at any one time. Using the atexit handler
with unique calls for each should be sufficient, rather than managing a list.
"""

import asyncio
import atexit
from typing import Union

from bleak import BleakClient
from bleak.backends.device import BLEDevice

import pyDE1
from pyDE1.btcontrack import persist_connection_file, remove_connection_file

logger = pyDE1.getLogger('Bluetooth.BleakClientWrapper')


class BleakClientWrapped (BleakClient):
    """
    See also previous code in scale.py and de1.py

        def atexit_disconnect(self):
            def sync_disconnect():
    """
    def __init__(self, address_or_ble_device: Union[BLEDevice, str], **kwargs):
        super(BleakClientWrapped, self).__init__(
            address_or_ble_device=address_or_ble_device)
        self._willful_disconnect = False

    @property
    def name(self):
        try:
            retval = self._device_info['Name']
        except (KeyError, AttributeError, TypeError):
            # CoreBluetooth on bleak 0.11.0 and 0.12.0
            # TypeError if not connected
            retval = None
        return retval

    @property
    def willful_disconnect(self):
        return self._willful_disconnect


    def sync_disconnect(self):
        if not self.is_connected:
            logger.debug(
                f"atexit sync_disconnect: Not connected to {self.name} at "
                f"{self.address}")
            print(f"atexit sync_disconnect: Not connected to {self.name} at "
                f"{self.address}")
            return
        else:
            logger.info(
                f"atexit sync_disconnect: Disconnecting {self.name} at "
                f"{self.address}")
            print(f"atexit sync_disconnect: Disconnecting {self.name} at "
                f"{self.address}")
            loop = asyncio.get_event_loop()
            if not loop.is_running():
                logger.info("atexit sync_disconnect: Starting event loop")
                print(f"atexit sync_disconnect: Starting event loop")
                loop.run_until_complete(self.disconnect())
            else:
                logger.info("atexit sync_disconnect: Using running loop")
                print(f"atexit sync_disconnect: Using running loop")
                loop.create_task(self.disconnect())
            # else:
            #     logger.critical(
            #         f"atexit sync_disconnect: NO LOOP AVAILABLE. "
            #         f"Unable to disconnect {self.name} at {self.address}"
            #     )
            #     print(f"atexit sync_disconnect: NO LOOP AVAILABLE. "
            #         f"Unable to disconnect {self.name} at {self.address}")

    async def connect(self, **kwargs) -> bool:
        persist_connection_file(self.address)
        atexit.register(self.sync_disconnect)
        retval = await super(BleakClientWrapped, self).connect(**kwargs)
        if retval:
            self._willful_disconnect = False
        return retval

    async def disconnect(self) -> bool:
        self._willful_disconnect = True
        retval = await super(BleakClientWrapped, self).disconnect()
        self._willful_disconnect = False
        if retval:
            logger.debug("Unregistering atexit disconnect "
                         f"{self.name} at {self.address}")
            atexit.unregister(self.sync_disconnect)
            remove_connection_file(self.address)
        return retval







"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only

"Main" executable:
  * Set things up
  * Start and supervise processes
  * Manage aggregated logging
  * Manage top-level shutdown
"""

# Supervise:
#   Task: log_queue_reader_blocks
#   Processes: Controller, OutboundAPI, InboundAPI

import pyDE1
import pyDE1.config
import pyDE1.pyde1_logging as pyde1_logging

from pyDE1.config import config


def run():
    # NB: Can only be set once, make sure the top-level script uses
    # if __name__ == '__main__':
    #   Not enough if importing something that imports multiprocessing
    #   to avoid RuntimeError: context has already been set
    # TODO: Replace this rather ugly hack on set_start_method, if possible

    import multiprocessing
    multiprocessing.set_start_method('spawn', force=True)

    import asyncio
    import atexit
    import logging
    import os
    import signal
    import time
    from types import FrameType

    import pyDE1.shutdown_manager as sm
    from pyDE1.api.outbound.mqtt import run_mqtt_outbound, OutboundMode
    from pyDE1.api.inbound.http import run_api_inbound
    from pyDE1.controller import run_controller
    from pyDE1.database.run import run_database_logger
    from pyDE1.database.manage import check_schema
    from pyDE1.supervise import SupervisedProcess

    logger = pyDE1.getLogger('Run')

    log_queue = multiprocessing.Queue()
    log_mqtt_pipe_read, log_mqtt_pipe_write = multiprocessing.Pipe(
        duplex=False)
    pyde1_logging.setup_queue_and_listener(config.logging,
                                           log_queue,
                                           log_mqtt_pipe_write)
    pyde1_logging.setup_queue_logging(config.logging, log_queue)
    pyde1_logging.config_logger_levels(config.logging)

    loop = asyncio.get_event_loop()
    loop.set_debug(True)

    def _sigchild_handler(signum: signal.Signals, frame: FrameType):
        ac = multiprocessing.active_children()
        logger.debug(
            f"Active children: {ac}")
        if sm.shutdown_underway.is_set() and len(ac) == 0:
            logger.debug("Setting cleanup_complete")
            sm.cleanup_complete.set()

    signal.signal(signal.SIGCHLD, _sigchild_handler)

    # Add as defined, to prevent NameError from a static list
    # and an early termination
    supervised_process_set = set()

    def on_shutdown_underway_cleanup():
        sm.shutdown_underway.wait()

        # This is intentional as the .do_not_restart setter
        # modifies the loop's readers
        def _the_rest_sync():
            logger.info("Setting do_not_restart")
            for sp in supervised_process_set:
                try:
                    sp.do_not_restart = True
                except AttributeError:
                    pass

            if sm.signal_rcvd is None:
                sig = signal.SIGTERM
            else:
                sig = sm.signal_rcvd
            for child in multiprocessing.active_children():
                # sp.terminate() would work, but pass the signal received
                logger.debug(f"os.kill {sig.name} {child.name}")
                os.kill(child.pid, sig)

        loop.call_soon_threadsafe(_the_rest_sync)

    on_shutdown_wait_task = loop.run_in_executor(
        None, on_shutdown_underway_cleanup)

    sm.attach_signal_handler_to_loop(sm.shutdown, loop)

    loop.set_exception_handler(sm.exception_handler)

    @atexit.register
    def kill_stragglers():
        procs = multiprocessing.active_children()
        if len(procs):
            print(f"kill {len(procs)} stragglers")
            for p in procs:
                print(f"Killing {p}")
                p.kill()
            print("buh-bye!")

    check_schema(loop)

    inbound_pipe_controller, inbound_pipe_server = multiprocessing.Pipe()

    # read, write, for simplex
    outbound_pipe_read, outbound_pipe_write = multiprocessing.Pipe(
        duplex=False)

    # MQTT logging
    supervised_outbound_log_process = SupervisedProcess(
        target=run_mqtt_outbound,
        kwargs={
            'config': config,
            'log_queue': log_queue,
            'outbound_pipe': log_mqtt_pipe_read,
            'mode': OutboundMode.LogRecord,
        },
        name='LogMQTT',
        daemon=False)
    supervised_process_set.add(supervised_outbound_log_process)
    supervised_outbound_log_process.start()

    # MQTT API
    supervised_outbound_api_process = SupervisedProcess(
        target=run_mqtt_outbound,
        kwargs={
            'config': config,
            'log_queue': log_queue,
            'outbound_pipe': outbound_pipe_read,
            'mode': OutboundMode.EventPayload,
        },
        name='OutboundAPI',
        daemon=False)
    supervised_process_set.add(supervised_outbound_api_process)
    supervised_outbound_api_process.start()

    # HTTP API
    supervised_inbound_api_process = SupervisedProcess(
        target=run_api_inbound,
        kwargs={
            'config': config,
            'log_queue': log_queue,
            'api_pipe': inbound_pipe_server,
        },
        name='InboundAPI',
        daemon=False)
    supervised_process_set.add(supervised_inbound_api_process)
    supervised_inbound_api_process.start()

    # 20 packets per second, 20 seconds ~ 400
    database_queue = multiprocessing.Queue(maxsize=400)

    # Database logging
    supervised_database_logger_process = SupervisedProcess(
        target=run_database_logger,
        kwargs={
            'config': config,
            'log_queue': log_queue,
            'notification_queue': database_queue,
        },
        name='DatabaseLogger',
        daemon=False)
    supervised_process_set.add(supervised_database_logger_process)
    supervised_database_logger_process.start()

    # Core logic

    # TODO: Not clear how this should restart
    #       as the DE1 will need to be reinitialized
    supervised_controller_process = SupervisedProcess(
        target=run_controller,
        kwargs={
            'config': config,
            'log_queue': log_queue,
            'inbound_pipe': inbound_pipe_controller,
            'outbound_pipe': outbound_pipe_write,
            'database_queue': database_queue,
        },
        name="Controller",
        daemon=False
    )
    supervised_process_set.add(supervised_controller_process)
    supervised_controller_process.start()

    logger.info('About to start loop')

    loop.run_forever()

    logger.debug("After loop.run_forever()")
    # explicit TPE (thread pool executor) shutdown hangs
    # print("shutdown TPE")
    # logging_tpe.shutdown(cancel_futures=True)
    # print("after shutdown TPE")
    ac = multiprocessing.active_children()
    if len(ac):
        level = logging.ERROR
    else:
        level = logging.DEBUG
    logger.log(level, f"Active_children: {multiprocessing.active_children()}")

    # loop.close()
    #
    # # loop.close() seems to be the source of a kill-related exit code
    # logger.debug("After loop.close()")

    procs = multiprocessing.active_children()
    if len(procs):
        logger.error(f"Need to kill {len(procs)} stragglers")
        for p in procs:
            logger.warning(f"Killing {p}")
            p.kill()

    ev_str = ''
    try:
        ev_str = signal.Signals(-sm.exit_value).name
    except ValueError:
        if sm.exit_value == os.EX_SOFTWARE:
            ev_str = 'os.EX_SOFTWARE'

    logger.info(f"Will exit with {sm.exit_value} {ev_str}")
    pyde1_logging.log_queue_listener.stop()
    # Thread needs a bit to shut down
    # TODO: Can/should this thread be joined?
    time.sleep(1)
    exit(sm.exit_value)


if __name__ == "__main__":

    # Complains even if set here
    # multiprocessing.set_start_method('spawn')

    import argparse

    pyde1_logging.setup_initial_logger()

    ap = argparse.ArgumentParser(
        description="""Main executable to start the pyDE1 core.

        """
        f"Default configuration file is at {pyDE1.config.DEFAULT_CONFIG_FILE}"
    )
    ap.add_argument('-c', type=str, help='Use as alternate config file')
    ap.add_argument('--console', action='store_true',
                    help='Timestamped, DEBUG level on stderr logging')

    args = ap.parse_args()

    config.load_from_yaml(args.c)

    if args.console:
        config.logging.handlers.STDERR = 'DEBUG'
        config.logging.formatters.STDERR = config.logging.formatters.LOGFILE

    run()



"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""
import collections
import json

"""
Logging Objectives:

  * The location and verbosity of the log file can be easily configured

  * The log file contains enough information to be useful to understand
    how the action progressed and why something might have gone wrong
    
  * The output to stderr, when run as a service, is meaningful to confirm
    that the service has started properly, as well as to capture any
    run-time errors
    
  * The output to stderr cam be easily configured for interactive debugging,
    including timestamps (omitted when running as a service, as they are 
    timestamped by systemd or the system logging system) and increasing
    verbosity as needed
    
  * The logging verbosity of the pyDE1 code can be adjusted, as needed,
    on a fine-grained basis
    
  * The logging verbosity of third-party libraries can be adjusted, as needed
  
  * Log-entry sequence is preserved no matter from which process it originates
  
  * Consistent initialization and configuration across processes


Python logging has a few quirks:
  
  * Entry of a log message to the logging system is gated by the effective
    level of the logger. That may be a level set on the logger itself,
    or, if none is set, traversing upwards through the parents until
    one with a level explicitly set is found. 
    
  * Exit of a log message that is in the system already is determined by
    the level set on the handler (not that of the logger to which it is 
    attached). When using a QueueListener, it needs to be instantiated with
    respect_handler_level=True for this behavior to occur.
    
  * LogRecord does not need to be formatted by the QueueHandler,
    it can be formatted by the "delivery" handler
    
  * dictConfig() apparently can't manage the QueueListener's handlers directly
  
  * Calling the module-level logging.info() or the like implicitly calls
      if len(root.handlers) == 0:
        basicConfig()
  
High-Level Design:

  * Immediately initialize a stderr-based logger that can be used until
    configuration can be read and parsed, then the "real" logging structures 
    created and configured
    
  * All delivery is through handlers on a QueueListener
    * Runs in the main process
      * First process up, hopefully the last down
      * The main process doesn't do much else
    * The QueueListener and its handlers are available in this module
      (for the main process, they are None here in other processes)
    * By setting the formatter and level on the handler (or filters, or ...)
      the output can be controlled from what input was let into the system

  * Leave the root logger with only a "plain" QueueHandler
    * This blocks inadvertent calls to logging.basicConfig()
    
  * Move all pYDE1-related logging to be children of 'pyde1' (WIP)
    * This allows setting to DEBUG for effective-level inheritance
      without setting third-party packages' effective level
      
  * Use a separate logging config file that will allow setting of 
    specific logging levels on the logger, to control entry into 
    the logging system. The separate file can be dropped in 
    without changing the general config file. By using dictConfig(), 
    it should be possible to configure the loggers' levels, 
    pyde1 or third-party, without having to enumerate them in advance.
    
    
Note: Can't 'from import pyDE1.config import config' 
      as may not be the right config if not pyDE1 "main"
"""

# Get all instantiated loggers: logging.root.manager.loggerDict

import copy
import logging
import multiprocessing
import multiprocessing.connection as mpc
import os
import warnings
from logging.handlers import QueueHandler, MemoryHandler
from typing import Optional, Union

import pyDE1
from pyDE1.config_load import ConfigLoadable


def setup_initial_logger():
    """
    Configure a basic logger that logs to stderr at DEBUG level
    sufficient to report errors in reading config and then
    starting the "real" logger.

    Capture the records in a MemoryHandler for later flushing
    to the logfile_handler, when it becomes available
    """
    CAPACITY = 1000  # records

    global memory_handler

    formatter = Formatter(
        "%(asctime)s %(levelname)s [%(processName)s] %(name)s: %(message)s")

    initial_handler = logging.StreamHandler()
    initial_handler.setFormatter(formatter)
    initial_handler.setLevel(logging.DEBUG)
    initial_handler.name = 'initial_logger'

    memory_handler = MemoryHandlerMulti(
        capacity=CAPACITY,
        flushLevel=(logging.CRITICAL + 1),
        target = None,
        flushOnClose=True
    )
    memory_handler.name = 'memory_handler'

    root_logger = logging.getLogger()
    if len(rlh := root_logger.handlers):
        root_logger.warning(
            f"setup_initial_logger found exising handlers {rlh}, removing"
        )
        for handler in rlh:
            root_logger.removeHandler(handler)
    root_logger.addHandler(initial_handler)
    root_logger.addHandler(memory_handler)
    root_logger.setLevel(logging.DEBUG)


class Formatter (logging.Formatter):
    """
    Change the logger name on output to remove _ROOT_LOGGER_PREFIX = 'pyDE1'
    and add 'root.' to others
    """

    # Same init, just override the format calls that reference the name

    @classmethod
    def revised_record(cls, record: logging.LogRecord) -> logging.LogRecord:
        revised = copy.copy(record)
        name = revised.name
        if name.startswith(pyDE1._ROOT_LOGGER_PREFIX):
            revised.name = name[pyDE1._ROOT_LOGGER_PREFIX_LEN:]
        elif name == 'root':
            pass
        else:
            revised.name = 'root.' + name
        return revised

    def format(self, record: logging.LogRecord) -> str:
        return super(Formatter, self).format(
            self.revised_record(record))

    # Don't override both, or record gets converted twice
    # def formatMessage(self, record: logging.LogRecord) -> str:
    #     return super(Formatter, self).formatMessage(
    #         self.revised_record(record))


# Common definition of the 'logging' section of config

class ConfigLogging (ConfigLoadable):
    def __init__(self):
        self.LOG_DIRECTORY = '/var/log/pyde1/'
        # NB: The log file name is matched against [a-zA-Z0-9._-]
        self.LOG_FILENAME = None    # Needs to be overridden
        self.formatters = ConfigLoggingFormatters()
        self.handlers = ConfigLoggingHandlers()
        self.LOGGERS = {
            'root.aiosqlite':   'INFO',
            'root.asyncio':     'INFO',
            'root.bleak':       'INFO',
        }


class ConfigLoggingFormatters (ConfigLoadable):
    def __init__(self):
        self.STYLE = '%'
        self.LOGFILE = '%(asctime)s ' \
                       '%(levelname)s [%(processName)s] %(name)s: %(message)s'
        self.MQTT = self.LOGFILE
        self.STDERR  = '%(levelname)s [%(processName)s] %(name)s: %(message)s'


class ConfigLoggingHandlers (ConfigLoadable):
    def __init__(self):
        self.LOGFILE = 'INFO'
        self.MQTT  = 'ERROR'
        self.STDERR  = 'WARNING'


# In a module, unfortunately,
#     thing = property(thing_getter, thing_setter)
# just accesses the property object, rather than providing property behavior

log_queue_listener = None
memory_handler = None
stderr_handler = None
mqtt_handler = None
logfile_handler = None


def setup_queue_and_listener(config_logging: ConfigLogging, 
                             log_queue: multiprocessing.Queue,
                             mqtt_connection: mpc.Connection):
    """
    Call only from the process that will be doing the logging

    NB: Call *after* config has been read
    :param config_logging: 
    """
    _setup_logging_internal(config_logging,
                            log_queue=log_queue,
                            mqtt_connection=mqtt_connection)


def setup_direct_logging(config_logging: ConfigLogging):
    """
    Intended for single-process applications, such as Visualizer uploader
    
    NB: Call *after* config has been read
    :param config_logging: 
    """
    if multiprocessing.parent_process() is not None \
            or len(multiprocessing.active_children()) != 0:
        warnings.warn(
            "setup_direct_logging() called in a multiprocessing environment",
            RuntimeWarning
        )
    _setup_logging_internal(config_logging,
                            log_queue=None,
                            mqtt_connection=None)


def _setup_logging_internal(config_logging: ConfigLogging,
                            log_queue: Optional[multiprocessing.Queue],
                            mqtt_connection: Optional[mpc.Connection]):
    """
    If log_queue is present, set up queue and listener
    If None, set up direct logging
    :param config_logging: 
    """
    global log_queue_listener, \
        memory_handler, stderr_handler, mqtt_handler, logfile_handler

    # log_queue = multiprocessing.Queue()

    root_logger = logging.getLogger()

    stderr_handler = logging.StreamHandler()
    stderr_formatter = Formatter(fmt=config_logging.formatters.STDERR)
    stderr_handler.setFormatter(stderr_formatter)
    stderr_handler.setLevel(config_logging.handlers.STDERR)
    stderr_handler.name = 'stderr_handler'

    root_logger.info(
        f"Configured stderr_handler: {stderr_handler}")

    if mqtt_connection:
        mqtt_handler = PipeHandler(pipe_connection=mqtt_connection)
        mqtt_formatter = Formatter(fmt=config_logging.formatters.MQTT)
        mqtt_handler.setFormatter(mqtt_formatter)
        mqtt_handler.setLevel(config_logging.handlers.MQTT)
    else:
        mqtt_handler = logging.NullHandler()
    mqtt_handler.name = 'mqtt_handler'

    root_logger.info(
        f"Configured mqtt_handler: {mqtt_handler}")

    if not os.path.exists(config_logging.LOG_DIRECTORY):
        root_logger.error(
            "logfile_directory '{}' does not exist. Creating.".format(
                os.path.realpath(config_logging.LOG_DIRECTORY)
            )
        )
        # Will create intermediate directories
        # Does not use "mode" on intermediates
        os.makedirs(config_logging.LOG_DIRECTORY)

    fq_logfile = os.path.join(config_logging.LOG_DIRECTORY,
                              config_logging.LOG_FILENAME)

    logfile_handler = logging.handlers.WatchedFileHandler(fq_logfile)
    logfile_formatter = Formatter(fmt=config_logging.formatters.LOGFILE)
    logfile_handler.setFormatter(logfile_formatter)
    logfile_handler.setLevel(config_logging.handlers.LOGFILE)
    logfile_handler.name = 'logfile_handler'

    root_logger.info(
        f"Configured logfile_handler: {logfile_handler}")

    if log_queue is not None:

        log_queue_listener = logging.handlers.QueueListener(
            log_queue,
            stderr_handler,
            mqtt_handler,
            logfile_handler,
            respect_handler_level=True
        )
        log_queue_listener.start()
    
        root_logger.info(f"Started {log_queue_listener}")
        root_logger.debug(
            f"log_queue_listener handlers: {log_queue_listener.handlers}")

    else:
        
        for handler in root_logger.handlers:
            root_logger.removeHandler(handler)
        root_logger.addHandler(stderr_handler)
        # TODO: Is there a use case and meaningful way to add the mqtt_handler?
        # root_logger.addHandler(mqtt_handler)
        root_logger.addHandler(logfile_handler)

    if memory_handler:
        root_logger.removeHandler(memory_handler)
        memory_handler.target = (logfile_handler, mqtt_handler)
        memory_handler.close()
        memory_handler = None

    set_root_logger_levels(config_logging)


def get_int_from_level(logging_level: Union[int, str]):
    if isinstance(logging_level, str):
        # Despite the name, this will return a number from a string
        return logging.getLevelName(logging_level)
    else:
        return logging_level


def set_root_logger_levels(config_logging: ConfigLogging):
    # No easy way to "DRY" this that I can think of today
    try:
        level_stderr = stderr_handler.level
    except AttributeError:
        level_stderr = get_int_from_level(config_logging.handlers.STDERR)
    try:
        level_mqtt = mqtt_handler.level
    except AttributeError:
        level_mqtt = get_int_from_level(config_logging.handlers.STDERR)
    try:
        level_logfile = logfile_handler.level
    except AttributeError:
        level_logfile = get_int_from_level(config_logging.handlers.LOGFILE)
    level = min(level_logfile, level_mqtt, level_stderr)
    pyDE1.getLogger().setLevel(level)
    pyDE1.getLogger('root').setLevel(level)


def setup_queue_logging(config_logging: ConfigLogging,
                        queue: multiprocessing.Queue):
    """
    Convert root logger to just using a QueueHandler(queue) for everything

    It should be the case that only third-party logging is inheriting
    the root logger's level. Allow it to be set elsewhere. As noted,
    the level of the root logger only impacts ingress to the logging system.
    Do not change the root logger's level here.
    :param config_logging:
    """

    # Configure the QueueHandler first, so that messages aren't lost

    queue_handler = QueueHandler(queue)
    queue_handler.name = 'queue_handler'
    root_logger = logging.getLogger()
    root_logger.addHandler(queue_handler)

    # Remove other root handlers

    for handler in root_logger.handlers:
        if handler.name != 'queue_handler':
            root_logger.removeHandler(handler)

    set_root_logger_levels(config_logging)


def config_logger_levels(config_logging: ConfigLogging):
    set_root_logger_levels(config_logging)
    logger = pyDE1.getLogger('Logging.Config')
    for logger_name, logger_level in config_logging.LOGGERS.items():
        logger.info(f"Setting {logger_name} to {logger_level}")
        pyDE1.getLogger(logger_name).setLevel(logger_level)


def log_record_to_json(record: logging.LogRecord):
    """
    Used by the MQTT process to send an easily parsable representation
    in addition to the formatted string

    NB: message -- "The logged message, computed as msg % args.
                    This is set when Formatter.format() is invoked."
    """
    to_send = {'version': '1.0.0'}
    for attr in (
            'created',
            'levelname',
            'levelno',
            'message',
            'name',
            'process',
            'processName',
            'thread',
            'threadName'
    ):
        to_send[attr] = record.__getattribute__(attr)
    return json.dumps(to_send)


class PipeHandler (QueueHandler):
    """
    Just like a QueueHandler, except it uses a multiprocessing.Pipe's
    connection to .send() to instead of Queue.put_nowait()
    """

    def __init__(self, pipe_connection: mpc.Connection):
        logging.Handler.__init__(self)
        self.pipe_connection = pipe_connection

    def enqueue(self, record):
        self.pipe_connection.send(record)


class MemoryHandlerMulti (MemoryHandler):
    """
    Like a MemoryHandler, but uses an iterable list of targets
    """

    def flush(self):
        self.acquire()
        if self.target is None:
            return
        if isinstance(self.target, logging.Handler):
            super(MemoryHandlerMulti, self).flush()
        else:
            try:
                for this_target in self.target:
                    for record in self.buffer:
                        this_target.handle(record)
                self.buffer.clear()
            finally:
                self.release()


"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""

import asyncio
import time
from typing import Optional, List, Callable
from uuid import UUID

import pyDE1
from pyDE1.config import config
from pyDE1.de1 import DE1
from pyDE1.de1.c_api import API_MachineStates
from pyDE1.de1.events import StateUpdate
from pyDE1.dispatcher.resource import ConnectivityEnum
from pyDE1.event_manager import SubscribedEvent
from pyDE1.event_manager.events import ConnectivityState
from pyDE1.exceptions import DE1NoAddressError, DE1APIValueError
from pyDE1.scale import Scale, scale_factory, recognized_scale_prefixes
from pyDE1.scale.events import (
    ScaleWeightUpdate, ScaleTareSeen, WeightAndFlowUpdate
)
from pyDE1.scanner import (
    BleakScannerWrapped, DiscoveredDevices, find_first_matching
)
from pyDE1.singleton import Singleton

logger = pyDE1.getLogger('Scale.Processor')


class ScaleProcessor (Singleton):
    """
    Subscribes to weight-update events from a scale
    Provides estimates of weight and mass-flow via
    WeightAndFlowUpdate events

    Should be able to init and "wire to" without a scale
    Scales should be able to be changed on the fly
    """

    # NB: This is intentionally done in _singleton_init() and not __init__()
    #     See Singleton and Guido's notes there
    #
    # def __init__(self):
    #     pass

    def _singleton_init(self):

        self._scale: Optional[Scale] = None
        self._scale_weight_update_id: Optional[UUID] = None
        self._scale_tare_seen_id: Optional[UUID] = None
        self._state_update_id: Optional[UUID] = None
        self._history_max = 10  # Will be extended if needed by Estimator
        self._history_time: List[float] = []
        self._history_weight: List[float] = []
        self._history_lock = asyncio.Lock()
        # set_scale needs _history_lock
        self._scale = None
        self._event_weight_and_flow_update = SubscribedEvent(self)

        # init of Estimator checks that the targets are present
        # (good practice to explicily declare anyways)
        self._current_weight: float = 0
        self._current_weight_time: float = 0
        self._average_flow: float = 0
        self._average_flow_time: float = 0
        self._median_weight: float = 0
        self._median_weight_time: float = 0
        self._median_flow: float = 0
        self._median_flow_time: float = 0
        self._estimators = [
            CurrentWeight(self, '_current_weight'),
            AverageFlow(self, '_average_flow', 11),
            MedianWeight(self, '_median_weight', 11),
            MedianFlow(self, '_median_flow', 11, 5),
        ]

        # This is a bit tricky, as self._scale is None
        bare_scale = Scale()
        asyncio.create_task(
            bare_scale._event_connectivity.publish(
                bare_scale._connectivity_change(
                    arrival_time=time.time(),
                    state=ConnectivityState.DISCONNECTED))
        )

    @property
    def scale(self):
        return self._scale

    async def set_scale(self, scale: Optional[Scale]):
        # NB: A Scale is self-referential and may not be GCed if "dropped" here
        if self._scale == scale:
            return
        # Unsubscribe the existing scale
        if self._scale is not None:
            await asyncio.gather(
                self._scale.event_weight_update.unsubscribe(
                    self._scale_weight_update_id
                ),
                self._scale.event_tare_seen.unsubscribe(
                    self._scale_tare_seen_id
                ),
                DE1().event_state_update.unsubscribe(
                    self._state_update_id
                )
            )

        # Always set
        self._scale = scale

        if self._scale is not None:
            (
                self._scale_weight_update_id,
                self._scale_tare_seen_id,
                self._state_update_id,

            ) = await asyncio.gather(
                self._scale.event_weight_update.subscribe(
                    self._create_scale_weight_update_subscriber()),

                self._scale.event_tare_seen.subscribe(
                    self._create_scale_tare_seen_subscriber()),

                DE1().event_state_update.subscribe(
                    self._create_state_update_subscriber()),

                return_exceptions=True
            )
        await self._reset()  # New scale, toss old history

        return self

    # Provide "null-safe" methods for API

    @property
    def scale_address(self):
        if self.scale is not None:
            return self.scale.address
        else:
            return None

    @property
    def scale_name(self):
        if self.scale is not None:
            return self.scale.name
        else:
            return None

    @property
    def scale_type(self):
        if self.scale is not None:
            return self.scale.type
        else:
            return None

    @property
    def scale_connectivity(self):
        if self.scale is not None:
            return self.scale.connectivity
        else:
            return ConnectivityEnum.NOT_CONNECTED

    #
    # Self-contained call for API
    #

    async def first_if_found(self, doit: bool):
        if self.scale and self.scale.is_connected:
            logger.warning(
                "first_if_found requested, but already connected. "
                "No action taken.")
        elif not doit:
            logger.warning(
                "first_if_found requested, but not True. No action taken.")
        else:
            device = await find_first_matching(
                recognized_scale_prefixes())
            if device:
                await self.change_scale_to_id(device.address)
        return self.scale_address

    async def change_scale_to_id(self, ble_device_id: Optional[str]):
        """
        For now, this won't return until connected or fails to connect
        As a result, will trigger the timeout on API calls
        """
        logger.info(f"Request to replace scale with {ble_device_id}")

        # TODO: Need to make distasteful assumption that the id is the address
        try:
            if self.scale_address == ble_device_id:
                logger.info(f"Already using {ble_device_id}. No action taken")
                return
        except AttributeError:
            pass

        old_scale = self.scale

        try:
            if ble_device_id is None:
                # Straightforward request to disconnect and not replace
                await self.set_scale(None)

            else:
                ble_device = DiscoveredDevices().ble_device_from_id(ble_device_id)
                if ble_device is None:
                    logger.warning(f"No record of {ble_device_id}, initiating scan")
                    # TODO: find_device_by_filter doesn't add to DiscoveredDevices
                    ble_device = await BleakScannerWrapped.find_device_by_address(
                        ble_device_id,
                        timeout=config.bluetooth.CONNECT_TIMEOUT)
                if ble_device is None:
                    raise DE1NoAddressError(
                        f"Unable to find device with id: '{ble_device_id}'")

                try:
                    logger.info(f"Disconnecting {self.scale}")
                    await self.scale.disconnect()
                except AttributeError:
                    pass

                new_scale = scale_factory(ble_device)
                if new_scale is None:
                    raise DE1APIValueError(
                        f"No scale could be created from {ble_device}")

                await self.set_scale(new_scale)
                await self.scale.connect()

        finally:
            # Scale is self-referential, make sure GC-ed
            if old_scale is not None and self.scale != old_scale:
                await old_scale.disconnect()
                old_scale.decommission()
                del old_scale

    async def connectivity_setter(self, value):
        assert isinstance(value, ConnectivityEnum), \
            f"mode of {value} not a ConnectivityEnum"
        if value is ConnectivityEnum.NOT_CONNECTED:
            if self.scale is not None:
                await self.scale.disconnect()
        elif value is ConnectivityEnum.CONNECTED:
            if self.scale is not None:
                await self.scale.connect()
        else:
            raise DE1APIValueError(
                "Only CONNECTED and NOT_CONNECTED can be set, "
                f"not {value}")

    @property
    def event_weight_and_flow_update(self):
        return self._event_weight_and_flow_update

    async def _reset(self):
        async with self._history_lock:
            self._reset_have_lock()

    def _reset_have_lock(self):
        self._history_time = []
        self._history_weight = []
        # TODO: Perhaps should clear any pending updates
        #       as they may be pre-tare

    @property
    def _history_available(self):
        # Ultra safe
        return min(len(self._history_weight), len(self._history_time))

    def _create_scale_tare_seen_subscriber(self) -> Callable:
        scale_processor = self

        async def scale_tare_seen_subscriber(sts: ScaleTareSeen):
            nonlocal scale_processor
            await scale_processor._reset()

        return scale_tare_seen_subscriber

    def _create_scale_weight_update_subscriber(self) -> Callable:
        scale_processor = self

        async def scale_weight_update_subscriber(swu: ScaleWeightUpdate):
            nonlocal scale_processor
            # TODO: This has the potential to get really messy
            #       Make sure this doesn't block other things
            # The Skale can return multiple updates in milliseconds
            # Is there any guarantee that they will be processed in order?
            # "Acquiring a lock is fair: the coroutine that proceeds
            #  will be the first coroutine that started waiting on the lock."
            async with self._history_lock:
                # Detect a gap in reporting being "too long"
                # (typically from a disconnect/reconnect)
                # A skip of three at 150 ms per update with the Skale II
                TOO_LONG = 0.8 # seconds
                try:
                    if ((dt := swu.scale_time
                               - scale_processor._history_time[-1]) > TOO_LONG):
                        logger.warning(
                            "Resetting scale due to gap in reports: "
                            f"{dt:0.3f} > {TOO_LONG} s")
                        scale_processor._reset_have_lock()
                        return
                except IndexError:
                    pass  # (No elements in the history list)
                scale_processor._history_time.append(swu.scale_time)
                scale_processor._history_weight.append(swu.weight)
                # Unlikely, but possibly the case that history_max shrunk
                while len(scale_processor._history_time) \
                        > scale_processor._history_max:
                    scale_processor._history_time.pop(0)
                while len(scale_processor._history_weight) \
                        > scale_processor._history_max:
                    scale_processor._history_weight.pop(0)

                # There's nothing here really parallelizable
                for estimator in self._estimators:
                    estimator.estimate()

                await self._event_weight_and_flow_update.publish(
                    WeightAndFlowUpdate(
                        arrival_time=swu.arrival_time,
                        scale_time=swu.scale_time,
                        current_weight=self._current_weight,
                        current_weight_time=self._current_weight_time,
                        average_flow=self._average_flow,
                        average_flow_time=self._average_flow_time,
                        median_weight=self._median_weight,
                        median_weight_time=self._median_weight_time,
                        median_flow=self._median_flow,
                        median_flow_time=self._median_flow_time,
                    )
                )

        return scale_weight_update_subscriber

    def _create_state_update_subscriber(self) -> Callable:
        scale_processor = self

        async def state_update_subscriber(su: StateUpdate):
            nonlocal scale_processor
            scale = scale_processor.scale
            if (su.previous_state == API_MachineStates.Sleep
                    and su.state != API_MachineStates.Sleep
                    and scale.address is not None
                    and not scale.is_connected):
                logger.info(
                    "Reconnecting to scale as DE1 left Sleep to "
                    f"{API_MachineStates(su.state).name}")
                scale._reset_reconnect()
                await scale_processor.scale._reconnect()

        return state_update_subscriber


# Prevent dreaded "circular import" problems
from pyDE1.scale.estimator import CurrentWeight, AverageFlow, \
    MedianWeight, MedianFlow



"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""

import asyncio
import multiprocessing
import re
import time
import traceback
from concurrent.futures import Executor, ThreadPoolExecutor
from typing import Union, Awaitable, Callable, Optional, Mapping

import pyDE1
import pyDE1.shutdown_manager as sm


re_us_char = re.compile('_\w')

def upcase_us_char_match(m: re.Match):
    return m.group()[-1].upper()

def camelcase_from_underscore(underscored: str):
    all_but_first = re.sub(re_us_char, upcase_us_char_match, underscored)
    return all_but_first[0].upper() + all_but_first[1:]


T_Work = Union[asyncio.Future, asyncio.Task, Awaitable]
#              ThreadPoolExec  create_task   generic executor


class SupervisedWork:

    def __init__(self, routine: Callable, *args, **kwargs):
        self._routine = routine
        self._args = args
        self._kwargs = kwargs
        try:
            self._name \
                = f"{routine.__self__.__class__.__name__}.{routine.__name__}"
        except AttributeError:
            self._name = routine.__name__
        self._logger = pyDE1.getLogger(f"Supervised.Work.{self._name}")
        self._work = None
        self._cancelled_error: Optional[asyncio.CancelledError] = None
        self._start_time_list = []
        self._restart_count_limit = 2   # No more than 2 restarts in
        self._restart_count_window = 20 # seconds

    def _too_many_restarts(self):
        retval = False
        if len(self._start_time_list) > self._restart_count_limit:
            n_ago = self._start_time_list[-(self._restart_count_limit + 1)]
            dt = time.time() - n_ago
            retval = dt < self._restart_count_window
        return retval

    def _record_start(self):
        # For now, just use a simple append
        self._start_time_list.append(time.time())
        # but check if it gets crazy long
        if len(self._start_time_list) > 2 * self._restart_count_limit:
            self._logger.warning(
                f"Start-time list seems long: {len(self._start_time_list)}")

    # Subclasses should self-start to be consistent with unsupervised calls
    def _start(self, task: Optional[T_Work] = None):
        if task is None:
            self._logger.info(f"Starting")
        else:
            exc = task.exception()
            if exc is None \
                    and not task.cancelled() \
                    and not self._cancelled_error:
                self._logger.info(
                    f"Exiting as completed without exception")
                return
            # asyncio.CancelledError == asyncio.exceptions.CancelledError
            # True
            elif self._cancelled_error:
                self._logger.info(
                    f"Exiting as cancelled with '{self._cancelled_error}'")
                return
            elif task.cancelled():
                self._logger.info(
                    f"Exiting as task.cancelled()")
                return
            else:
                self._logger.exception(
                    f"Task failed with exception:", exc_info=exc)
                if self._too_many_restarts():
                    self._logger.critical("Too many restarts, abandoning")
                    self._logger.critical("Calling shutdown")
                    loop = asyncio.get_event_loop()
                    if loop.is_running():
                        loop.create_task(sm.shutdown(None, loop))
                    else:
                        loop.run_until_complete(sm.shutdown(None, loop))
                    return
                self._logger.info(f"Restarting")

        self._work = self._create_work()
        self._record_start()
        self._work.add_done_callback(self._start)

        return self  # Allows sw = SupervisedWork(routine, arg_list).start()

    def _create_work(self) -> T_Work:
        raise NotImplementedError

    @property
    def work(self):
        return self._work


class SupervisedTask (SupervisedWork):

    def __init__(self, routine: Callable, *args, **kwargs):
        super(SupervisedTask, self).__init__(routine, *args, **kwargs)
        self._logger = pyDE1.getLogger(f"Supervised.Task.{self._name}")
        self._start()

    def _create_work(self) -> T_Work:
        loop = asyncio.get_event_loop()
        async def wrapped(st: SupervisedTask):
            inner_task = st._routine(*st._args, **st._kwargs)
            try:
                await inner_task
            except asyncio.CancelledError as e:
                st._logger.debug(
                    f"{e}: {inner_task}")
                st._cancelled_error = e

        work = loop.create_task(
            wrapped(self),
            name=camelcase_from_underscore(self._name)
        )
        return work


class SupervisedExecutor (SupervisedWork):
    """
    Matches signature of loop.run_in_executor()

    NB: This assumes that the Executor returns an Awaitable
        that supports .add_done_callback()

        concurrent.futures.ThreadPoolExecutor() does
        others are not tested at this time.

    From Python BaseEventLoop:

    def set_default_executor(self, executor):
        if not isinstance(executor, concurrent.futures.ThreadPoolExecutor):
        warnings.warn(
            'Using the default executor that is not an instance of '
            'ThreadPoolExecutor is deprecated and will be prohibited '
            'in Python 3.9',
            DeprecationWarning, 2)
    """
    def __init__(self, executor: [Executor],
                 routine: Callable, *args):
        super(SupervisedExecutor, self).__init__(routine, *args)
        self._logger = pyDE1.getLogger(f"Supervised.Executor.{self._name}")
        if executor is None:
            executor = ThreadPoolExecutor(
                max_workers=1,
                thread_name_prefix=camelcase_from_underscore(self._name))
        self._executor = executor
        self._start()

    def _create_work(self) -> T_Work:
        loop = asyncio.get_event_loop()
        work = loop.run_in_executor(self._executor,
                                    self._routine, *self._args)
        return work


class SupervisedProcess:
    """
    A multiprocessing.Process that will restart itself should it terminate

    Calling SupervisedProcess.terminate() or .kill() will prevent restart
    Setting .do_not_restart will also prevent restart

    Semantics as with multiprocessing.Process()
    Needs to be explicitly started as a result

    NB: Returns a SupervisedProcess object, not a Process object

    TODO: Custom __getarg__ to delegate to Process
    """

    def __init__(self, target: Callable, name: Optional[str] = None,
                 args: Optional[tuple] = (), kwargs: Optional[Mapping] = None,
                 daemon: Optional[bool] = None,
                 do_not_restart=False):
        self._target = target
        self._name = name
        self._args = args
        self._kwargs = kwargs
        self._daemon = daemon
        self._do_not_restart = do_not_restart
        self._logger = pyDE1.getLogger(f"Supervised.Process.{self._name}")
        self._process: Optional[multiprocessing.Process] = None
        self._start_time_list = []
        self._restart_count_limit = 2  # No more than 2 restarts in
        self._restart_count_window = 20  # seconds

    def _too_many_restarts(self):
        retval = False
        if len(self._start_time_list) > self._restart_count_limit:
            n_ago = self._start_time_list[-(self._restart_count_limit + 1)]
            dt = time.time() - n_ago
            retval = dt < self._restart_count_window
        return retval

    def _record_start(self):
        # For now, just use a simple append
        self._start_time_list.append(time.time())
        # but check if it gets crazy long
        if len(self._start_time_list) > 2 * self._restart_count_limit:
            self._logger.warning(
                f"Start-time list seems long: {len(self._start_time_list)}")

    def _wrap_target(self, *args, **kwargs):
        try:
            self._target(*args, **kwargs)
        except Exception as exc:
            tbe = traceback.TracebackException.from_exception(exc)
            self._logger.error(
                f"Shutting down on exception: {''.join(tbe.format())}")
            if (loop := asyncio.get_event_loop()).is_running():
                self._logger.debug(f"Have running loop: {loop}")
                loop.create_task(sm.shutdown(None, loop))
            else:
                self._logger.debug("No running loop")
                loop.run_until_complete(sm.shutdown(None, loop))

    def _create_process(self):
        self._process = multiprocessing.Process(
            target=self._wrap_target,
            name=self._name,
            args=self._args,
            kwargs=self._kwargs,
            daemon=self._daemon,
        )

    def start(self):

        if self._process is not None:
            if self._do_not_restart:
                self._logger.info(
                    "Not restarting as do_not_restart is set.")
                # Without dealing with the add_reader on sentinel, thrashes
                return

            if self._process.is_alive():
                self._logger.warning(
                    "Process is already running. "
                    f"Not restarting {self._process}")
                return

        # All reasons not to start eliminated

        if self._process is None:
            self._logger.info(f"Starting")
        else:
            if self._too_many_restarts():
                self._logger.critical("Too many restarts, abandoning")
                self.do_not_restart = True
                self._logger.critical("Calling shutdown")
                loop = asyncio.get_event_loop()
                if loop.is_running():
                    loop.create_task(sm.shutdown(None, loop))
                else:
                    loop.run_until_complete(sm.shutdown(None, loop))
                return
            else:
                self._logger.info(f"Restarting")

        self._create_process()
        self._process.start()
        self._record_start()
        self._logger.info(f"Started: {self._process}")

        if not self.do_not_restart:
            asyncio.get_event_loop().add_reader(self._process.sentinel,
                                                self.sentinel_cb)

    def sentinel_cb(self):
        asyncio.get_event_loop().remove_reader(self._process.sentinel)
        if self.do_not_restart:
            self._logger.info("Process exited, do-not-restart set")
        else:
            self._logger.error("Process exited")
            self.start()

    @property
    def process(self):
        return self._process

    @property
    def do_not_restart(self):
        return self._do_not_restart

    @do_not_restart.setter
    def do_not_restart(self, flag: bool):
        flag = bool(flag)
        if flag == self._do_not_restart:
            return

        if self._process is not None:
            if flag:
                asyncio.get_event_loop().remove_reader(self._process.sentinel)
            else:
                asyncio.get_event_loop().add_reader(self._process.sentinel,
                                                    self.start)
        self._do_not_restart = flag


"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only

Mix-in classes with method for loading from YAML file into config

    ConfigLoadable  Knows how to load itself and descendants from a dict
    ConfigYAMLLoad  ConfigLoadable with ability to load from a YAML file

Also, as config is often loaded very early in the start-up process,
ensure that the self.logger has at least a handler that outputs to stderr
so that config success/errors get logged somewhere.
"""

import inspect
import logging
from typing import Optional

import yaml

import pyDE1


class ConfigLoadable:
    """
    Supports loading from a dict

    NB: Load is done recursively, without loop detection
    """

    _logger = pyDE1.getLogger('ConfigLoadable')

    def load_from_dict(self, source: dict):
        self._load_from_dict_inner(source=source, target=self,
                                   breadcrumbs=None)

    def _load_from_dict_inner(self, source: dict, target: "ConfigLoadable",
                              breadcrumbs: Optional[str] = None):
        for key, val in source.items():
            # Try to protect against accidental overwrites or additions
            if not isinstance(key, str):
                self._logger.error(
                    f"Expected a string for a key, "
                    f"not {key}, skipping")
                continue

            if breadcrumbs:
                next_breadcrumbs = f"{breadcrumbs}.{key}"
            else:
                next_breadcrumbs = key

            if key.startswith('_'):
                self._logger.error(
                    f"Underscore names {breadcrumbs} not permitted, skipping")
                continue

            try:
                apply_val_to = target.__getattribute__(key)
            except AttributeError:
                self._logger.error(
                    f"No such key {next_breadcrumbs}, skipping")
                continue

            if inspect.ismethod(apply_val_to):
                self._logger.error(
                    f"Can't override methods at {next_breadcrumbs}, skipping")
                continue

            if isinstance(apply_val_to, ConfigLoadable):
                if not isinstance(val, dict):
                    if val is not None:
                        self._logger.error(
                            f"Expected a dict for {next_breadcrumbs}, "
                            f"not {type(val)}, skipping")
                    else:
                        self._logger.warning(
                            f"No entries found for {next_breadcrumbs}, "
                            "skipping (may be intentional)")
                    continue
                self._load_from_dict_inner(val, apply_val_to, next_breadcrumbs)
            else:
                self._logger.debug(f"Setting {next_breadcrumbs}")
                target.__setattr__(key, val)


class ConfigYAML (ConfigLoadable):

    # Override in implementations
    DEFAULT_CONFIG_FILE = None

    def ensure_stderr_handler(self):
        root_logger = logging.getLogger()
        self._logger = pyDE1.getLogger('Config.YAML')
        need_stderr_handler = True
        for handler in root_logger.handlers:
            try:
                if isinstance(handler, logging.StreamHandler) \
                        and handler.stream.name == '<stderr>':
                    need_stderr_handler = False
                    break
            except AttributeError:
                pass
        if need_stderr_handler:
            for handler in self._logger.handlers:
                try:
                    if isinstance(handler, logging.StreamHandler) \
                            and handler.stream.name == '<stderr>':
                        need_stderr_handler = False
                        break
                except AttributeError:
                    pass
        if need_stderr_handler:
            handler = logging.StreamHandler()
            handler.setLevel(logging.DEBUG)
            handler.setFormatter(logging.Formatter(
                "%(asctime)s %(levelname)s [%(processName)s] %(name)s: "
                "%(message)s"
            ))
            self._logger.addHandler(handler)
        self._logger.setLevel(logging.DEBUG)

    def load_from_yaml(self, filename: Optional[str] = None):
        self.ensure_stderr_handler()
        if filename is None:
            filename = self.DEFAULT_CONFIG_FILE
        if filename is None:
            self._logger.critical("No DEFAULT_CONFIG_FILE specified for "
                                 f"{self.__class__.__name__} exiting.")
            raise FileNotFoundError("Coding error - no DEFAULT_CONFIG_FILE")
        self._logger.info(f"Loading config overrides from {filename}")
        parsed = {}
        try:
            with open(file=filename, mode='r') as fh:
                parsed = yaml.safe_load(fh)
        except FileNotFoundError as e:
            if filename != self.DEFAULT_CONFIG_FILE:
                self._logger.critical(
                    f"Unable to open config file '{filename}', exiting.")
                raise e
            else:
                self._logger.warning(
                    f"Could not find default config file {self.DEFAULT_CONFIG_FILE}")
                return

        except Exception as e:
            self._logger.critical(
                f"Error parsing config from '{filename}', exiting.")
            raise e

        # self._old_loader(parsed)
        self.load_from_dict(parsed)

        self._logger.info(f"Config overrides loaded from {filename}")



"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only

Somewhat generic signal and exception handlers for "graceful" shutdown.

App exit will be -signal_number or os_EX_SOFTWARE on exception handled here.

See further:
    https://docs.python.org/3/library/asyncio-eventloop.html
       #asyncio.loop.set_exception_handler
    https://www.roguelynn.com/words/asyncio-exception-handling/


Example cleanup routine:

async def wait_then_cleanup(client: mqtt.Client):
    await loop.run_in_executor(None, shutdown_underway.wait)
    client.disconnect()
    client.loop_stop()
    cleanup_complete.set()

"""
import asyncio
import logging
import multiprocessing
import os
import pprint
import signal
import traceback
from typing import Callable, Optional, Coroutine, Iterable, Union

import pyDE1

# Set when a shutdown is requested
# Does not start a shutdown when set
shutdown_underway = multiprocessing.Event()

# Should be set by app-specific clean-up routines
cleanup_complete = multiprocessing.Event()

# Give up on waiting for cleanup_complete.is_set() after
CLEANUP_WAIT = 5.0  # seconds

exit_value = os.EX_OK
signal_rcvd = None

logger = pyDE1.getLogger('Shutdown')

default_signal_set = (
    signal.SIGHUP,
    signal.SIGINT,
    signal.SIGQUIT,
    signal.SIGABRT,
    signal.SIGTERM,
)


def task_to_string(t: asyncio.Task):
    return f"{t.get_name()}: {t.get_coro().cr_frame.f_code}"


def exception_handler(loop: asyncio.AbstractEventLoop,
                      context: dict):
    if shutdown_underway.is_set():
        level = logging.WARNING
    else:
        level = logging.CRITICAL
    exc_class = context['exception'].__class__
    logger.log(level,
        f"Uncaught exception (loop) {exc_class}:\n"
        f"{pprint.pformat(context)}")
    if not shutdown_underway.is_set():
        loop.create_task(shutdown(None, loop))


async def shutdown(sig: Optional[signal.Signals],
                   loop: asyncio.AbstractEventLoop):
    global exit_value
    global signal_rcvd
    signal_rcvd = sig

    if shutdown_underway.is_set():
        if sig is not None:
            name_str = f"from {sig.name}"
        else:
            name_str = " (no signal passed)"
        logger.info(f"Already shutting down, ignoring {name_str}")
        return

    shutdown_underway.set() # Cleanup should trigger off this
    graceful = (sig == signal.SIGTERM)

    if graceful: # The "expected" shutdown signal
        logger.info(f"Shutting down from {sig.name}")
    elif sig is None:
        # Can happen if called due to an uncaught exception
        logger.critical(
            "Shutdown called without signal (likely an uncaught exception)")
    else:
        logger.critical(f"Shutdown called due to {sig.name}")

    logger.info(f"Waiting up to {CLEANUP_WAIT} seconds for clean up.")
    signaled = await loop.run_in_executor(None,
                                          cleanup_complete.wait, CLEANUP_WAIT)
    if signaled:
        logger.info("cleanup_complete set, continuing")
    else:
        logger.error("cleanup_complete not set fast enough, continuing")

    # Thread exit requires a flag that is periodically checked
    # it is nearly impossible to "kill" a thread
    logger.info("Shutting down default executor")
    await loop.shutdown_default_executor()
    logger.info("Shutting down asyncgens")
    await loop.shutdown_asyncgens()

    me = asyncio.current_task(loop)
    tasks_to_cancel = [t for t in asyncio.all_tasks(loop)
                       if t is not me]
    logger.info(f"Cancelling {len(tasks_to_cancel)} running tasks")
    for t in tasks_to_cancel:
        logger.info(f"Cancelling {task_to_string(t)}")
        try:
            t.cancel("Shutdown underway")
        except asyncio.exceptions.CancelledError:
            pass
    await asyncio.gather(*tasks_to_cancel, return_exceptions=True)

    logger.info("Stopping asyncio loop")

    loop.stop()
    # .close() here raises RuntimeError('Cannot close a running event loop')
    # logger.info("Closing loop")
    # loop.close()
    if graceful:
        exit_value = os.EX_OK
    elif sig:
        exit_value = -sig.value
    else:
        exit_value = os.EX_SOFTWARE


def attach_signal_handler_to_loop(signal_handler: Callable[
                        [signal.Signals, asyncio.AbstractEventLoop], Coroutine],
                                  loop: asyncio.AbstractEventLoop,
                                  signals: Optional[Union[
                                      Iterable[signal.Signals],
                                      signal.Signals]] = default_signal_set):

    if isinstance(signals, signal.Signals):
        signals = (signals,)

    # See https://www.roguelynn.com/words/asyncio-graceful-shutdowns/
    #     https://docs.python-guide.org/writing/gotchas/#late-binding-closures
    # for why sig needs to be bound in the lambda
    for sig in signals:
        loop.add_signal_handler(
            sig,
            lambda s=sig: asyncio.create_task(signal_handler(s, loop))
        )


# Usually also needed:
#       loop.set_exception_handler(exception_handler)

# Useful as a callback on tasks
def shutdown_if_exception(fut: asyncio.Future):
    if fut.exception():
        tbe = traceback.TracebackException.from_exception(fut.exception())
        if shutdown_underway.is_set():
            level = logging.WARNING
        else:
            level = logging.CRITICAL
        logger.log(level,
                   "Uncaught exception (future): " + ''.join(tbe.format()))
        if not shutdown_underway.is_set():
            logger.critical("Initiating shutdown")
            loop = asyncio.get_running_loop()
            loop.create_task(shutdown(None, loop))



